from __future__ import annotations


__copyright__ = """
Copyright (C) 2012 Andreas Kloeckner
Copyright (C) 2022 University of Illinois Board of Trustees
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import logging
from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, Callable, Sequence


logger = logging.getLogger(__name__)

from constantdict import constantdict

import islpy as isl
from pytools import memoize_on_first_arg

from loopy.diagnostic import LoopyError, ReductionIsNotTriangularError, warn_with_kernel
from loopy.kernel.data import AddressSpace, TemporaryVariable, make_assignment
from loopy.kernel.function_interface import CallableKernel
from loopy.kernel.instruction import Assignment, InstructionBase, MultiAssignmentBase
from loopy.symbolic import ReductionCallbackMapper
from loopy.transform.instruction import replace_instruction_ids_in_insn
from loopy.translation_unit import ConcreteCallablesTable, TranslationUnit


if TYPE_CHECKING:
    from pymbolic.primitives import ExpressionNode
    from pytools.tag import Tag

    from loopy.kernel import LoopKernel


# {{{ reduction realization context

@dataclass
class _ChangeFlag:
    changes_made: bool


@dataclass(frozen=True)
class _ReductionRealizationContext:
    # {{{ read-only

    mapper: RealizeReductionCallbackMapper

    force_scan: bool
    automagic_scans_ok: bool
    unknown_types_ok: bool

    # FIXME: This feels like a broken-by-design concept.
    force_outer_iname_for_scan: str | None

    # We use the original kernel for a number of lookups whose value
    # we do not change and which might be already cached on it.
    orig_kernel: LoopKernel
    kernel: LoopKernel

    id_prefix: str

    # }}}

    # {{{ internally mutable, same across entire recursion

    insn_id_gen: Callable[[str], str]
    var_name_gen: Callable[[str], str]

    additional_temporary_variables: dict[str, TemporaryVariable]
    additional_insns: list[InstructionBase]
    domains: list[isl.BasicSet]
    additional_iname_tags: dict[str, Sequence[Tag]]
    # list only to facilitate mutation
    boxed_callables_table: list[ConcreteCallablesTable]

    # FIXME: This is a broken-by-design concept. Local-parallel scans emit a
    # reduction internally. This serves to avoid force_scan acting on that
    # reduction.
    inames_added_for_scan: set[str]

    # }}}

    # {{{ surrounding instruction, read-only (different at each recursive level)

    # These are attributes from 'surrounding' instruction, for generated
    # instructions to potentially inherit.
    surrounding_within_inames: frozenset[str]
    surrounding_depends_on: frozenset[str]
    surrounding_no_sync_with: frozenset[tuple[str, str]]
    surrounding_predicates: frozenset[ExpressionNode]

    # }}}

    # {{{ surrounding instruction, internally mutable
    # (different at each recursive level)

    # These are requested additions to attributes of the surrounding instruction.

    # FIXME add_within_inames seems broken by design.
    surrounding_insn_add_within_inames: set[str]

    surrounding_insn_add_depends_on: set[str]
    surrounding_insn_add_no_sync_with: set[tuple[str, str]]

    # }}}

    # {{{ change tracking (same across entire recursion)

    _change_flag: _ChangeFlag

    @property
    def were_changes_made(self):
        return self._change_flag.changes_made

    def changes_made(self):
        self._change_flag.changes_made = True

    # }}}

    def new_subinstruction(self, *, within_inames, depends_on,
            no_sync_with=None, predicates=None):
        if no_sync_with is None:
            no_sync_with = self.surrounding_no_sync_with
        if predicates is None:
            predicates = self.surrounding_predicates

        return replace(self,
                surrounding_within_inames=within_inames,
                surrounding_depends_on=depends_on,
                surrounding_no_sync_with=no_sync_with,
                surrounding_predicates=predicates,

                surrounding_insn_add_within_inames=set(),
                surrounding_insn_add_depends_on=set(),
                surrounding_insn_add_no_sync_with=set())

    def get_insn_kwargs(self):
        return {
                "within_inames": (
                    self.surrounding_within_inames
                    | frozenset(self.surrounding_insn_add_within_inames)),
                "within_inames_is_final": True,
                "depends_on": (
                    self.surrounding_depends_on
                    | frozenset(self.surrounding_insn_add_depends_on)),
                "no_sync_with": (
                    self.surrounding_no_sync_with
                    | frozenset(self.surrounding_insn_add_no_sync_with)),
                "predicates": self.surrounding_predicates}

# }}}


# {{{ iname/domain wrangling

@dataclass(frozen=True)
class _InameClassification:
    sequential: tuple[str, ...]
    local_parallel: tuple[str, ...]
    nonlocal_parallel: tuple[str, ...]


def _classify_reduction_inames(red_realize_ctx, inames):
    sequential = []
    local_par = []
    nonlocal_par = []

    from loopy.kernel.data import (
        ConcurrentTag,
        LocalInameTagBase,
        UnrolledIlpTag,
        UnrollTag,
        filter_iname_tags_by_type,
    )

    for iname in inames:
        try:
            iname_tags = red_realize_ctx.additional_iname_tags[iname]
        except KeyError:
            iname_tags = red_realize_ctx.kernel.iname_tags(iname)

        if filter_iname_tags_by_type(iname_tags, (UnrollTag, UnrolledIlpTag)):
            # These are nominally parallel, but we can live with
            # them as sequential.
            sequential.append(iname)

        elif filter_iname_tags_by_type(iname_tags, LocalInameTagBase):
            local_par.append(iname)

        elif filter_iname_tags_by_type(iname_tags, ConcurrentTag):
            nonlocal_par.append(iname)

        else:
            sequential.append(iname)

    return _InameClassification(
            tuple(sequential), tuple(local_par), tuple(nonlocal_par))


def _add_params_to_domain(domain, param_names):
    dim_type = isl.dim_type
    nparams_orig = domain.dim(dim_type.param)
    domain = domain.add_dims(dim_type.param, len(param_names))

    for param_idx, param_name in enumerate(param_names):
        domain = domain.set_dim_name(
                dim_type.param, param_idx + nparams_orig, param_name)

    return domain


def _move_set_to_param_dims_except(domain, except_dims):
    dim_type = isl.dim_type

    iname_idx = 0
    for iname in domain.get_var_names(dim_type.set):
        if iname not in except_dims:
            domain = domain.move_dims(
                    dim_type.param, 0,
                    dim_type.set, iname_idx, 1)
            iname_idx -= 1
        iname_idx += 1

    return domain


def _domain_depends_on_given_set_dims(domain, set_dim_names):
    set_dim_names = frozenset(set_dim_names)

    return any(
            set_dim_names & set(constr.get_coefficients_by_name())
            for constr in domain.get_constraints())


def _insert_subdomain_into_domain_tree(kernel, domains, subdomain):
    # Intersect with inames, because we could have captured some kernel params
    # in here too...
    dependent_inames = (
            frozenset(subdomain.get_var_names(isl.dim_type.param))
            & kernel.all_inames())
    idx, = kernel.get_leaf_domain_indices(dependent_inames)
    domains.insert(idx + 1, subdomain)

# }}}


# {{{ scan inference

def _check_reduction_is_triangular(kernel, expr, scan_param):
    """Check whether the reduction within `expr` with scan parameters described by
    the structure `scan_param` is triangular. This attempts to verify that the
    domain for the scan and sweep inames is as follows:

    [params] -> {
        [other inames..., scan_iname, sweep_iname]:
            (sweep_min_value
                <= sweep_iname
                <= sweep_max_value)
            and
            (scan_min_value
                <= scan_iname
                <= stride * (sweep_iname - sweep_min_value) + scan_min_value)
            and
            (irrelevant constraints)
    }
    """

    orig_domain = kernel.get_inames_domain(
            frozenset((scan_param.sweep_iname, scan_param.scan_iname)))

    sweep_iname = scan_param.sweep_iname
    scan_iname = scan_param.scan_iname
    affs = isl.affs_from_space(orig_domain.space)

    sweep_lower_bound = isl.align_spaces(
            scan_param.sweep_lower_bound,
            affs[0])

    sweep_upper_bound = isl.align_spaces(
            scan_param.sweep_upper_bound,
            affs[0])

    scan_lower_bound = isl.align_spaces(
            scan_param.scan_lower_bound,
            affs[0])

    from itertools import product

    for (sweep_lb_domain, sweep_lb_aff), \
        (sweep_ub_domain, sweep_ub_aff), \
        (scan_lb_domain, scan_lb_aff) in \
            product(sweep_lower_bound.get_pieces(),
                    sweep_upper_bound.get_pieces(),
                    scan_lower_bound.get_pieces()):

        # Assumptions inherited from the domains of the pwaffs
        assumptions = sweep_lb_domain & sweep_ub_domain & scan_lb_domain

        # Sweep iname constraints
        hyp_domain = affs[sweep_iname].ge_set(sweep_lb_aff)
        hyp_domain &= affs[sweep_iname].le_set(sweep_ub_aff)

        # Scan iname constraints
        hyp_domain &= affs[scan_iname].ge_set(scan_lb_aff)
        hyp_domain &= affs[scan_iname].le_set(
                scan_param.stride * (affs[sweep_iname] - sweep_lb_aff)
                + scan_lb_aff)

        hyp_domain, = (hyp_domain & assumptions).get_basic_sets()
        test_domain, = (orig_domain & assumptions).get_basic_sets()

        hyp_gist_against_test = hyp_domain.gist(test_domain)
        if _domain_depends_on_given_set_dims(hyp_gist_against_test,
                (sweep_iname, scan_iname)):
            return False, (
                    "gist of hypothesis against test domain "
                    "has sweep or scan dependent constraints: '%s'"
                    % hyp_gist_against_test)

        test_gist_against_hyp = test_domain.gist(hyp_domain)
        if _domain_depends_on_given_set_dims(test_gist_against_hyp,
                (sweep_iname, scan_iname)):
            return False, (
                   "gist of test against hypothesis domain "
                   "has sweep or scan dependent constraint: '%s'"
                   % test_gist_against_hyp)

    return True, "ok"


@dataclass(frozen=True)
class _ScanCandidateParameters:
    sweep_iname: str
    scan_iname: str
    sweep_lower_bound: isl.PwAff
    sweep_upper_bound: isl.PwAff
    scan_lower_bound: isl.PwAff
    stride: int


def _try_infer_scan_candidate_from_expr(
        kernel, expr, within_inames, sweep_iname=None):
    """Analyze `expr` and determine if it can be implemented as a scan.
    """
    from loopy.symbolic import Reduction
    assert isinstance(expr, Reduction)

    if len(expr.inames) != 1:
        raise ValueError(
                "Multiple inames in reduction: '{}'".format(", ".join(expr.inames)))

    scan_iname, = expr.inames

    from loopy.kernel.tools import DomainChanger
    dchg = DomainChanger(kernel, (scan_iname,))
    domain = dchg.get_original_domain()

    if sweep_iname is None:
        try:
            sweep_iname = _try_infer_sweep_iname(
                    domain, scan_iname, kernel.all_inames())
        except ValueError as v:
            raise ValueError(
                    "Couldn't determine a sweep iname for the scan "
                    "expression '%s': %s" % (expr, v)) from None

    try:
        sweep_lower_bound, sweep_upper_bound, scan_lower_bound = (
                _try_infer_scan_and_sweep_bounds(
                    kernel, scan_iname, sweep_iname, within_inames))
    except ValueError as v:
        raise ValueError(
                "Couldn't determine bounds for the scan with expression '%s' "
                "(sweep iname: '%s', scan iname: '%s'): %s"
                % (expr, sweep_iname, scan_iname, v)) from None

    try:
        stride = _try_infer_scan_stride(
                kernel, scan_iname, sweep_iname, sweep_lower_bound)
    except ValueError as v:
        raise ValueError(
                "Couldn't determine a scan stride for the scan with expression '%s' "
                "(sweep iname: '%s', scan iname: '%s'): %s"
                % (expr, sweep_iname, scan_iname, v)) from None

    return _ScanCandidateParameters(
            sweep_iname=sweep_iname,
            scan_iname=scan_iname,
            sweep_lower_bound=sweep_lower_bound,
            sweep_upper_bound=sweep_upper_bound,
            scan_lower_bound=scan_lower_bound,
            stride=stride)


def _try_infer_sweep_iname(domain, scan_iname, candidate_inames):
    """The sweep iname is the outer iname which guides the scan.

    E.g. for a domain of {[i,j]: 0<=i<n and 0<=j<=i}, i is the sweep iname.
    """
    constrs = domain.get_constraints()
    sweep_iname_candidate = None

    for constr in constrs:
        candidate_vars = {
                var for var in constr.get_var_dict()
                if var in candidate_inames}

        # Irrelevant constraint - skip
        if scan_iname not in candidate_vars:
            continue

        # No additional inames - skip
        if len(candidate_vars) == 1:
            continue

        candidate_vars.remove(scan_iname)

        # Depends on more than one iname - error
        if len(candidate_vars) > 1:
            raise ValueError(
                    "More than one sweep iname candidate for scan iname '%s' found "
                    "(via constraint '%s')" % (scan_iname, constr))

        next_candidate = candidate_vars.pop()

        if sweep_iname_candidate is None:
            sweep_iname_candidate = next_candidate
            defining_constraint = constr
        else:
            # Check next_candidate consistency
            if sweep_iname_candidate != next_candidate:
                raise ValueError(
                        "More than one sweep iname candidate for scan iname '%s' "
                        "found (via constraints '%s', '%s')" %
                        (scan_iname, defining_constraint, constr))

    if sweep_iname_candidate is None:
        raise ValueError(
                "Couldn't find any sweep iname candidates for "
                "scan iname '%s'" % scan_iname)

    return sweep_iname_candidate


def _try_infer_scan_and_sweep_bounds(kernel, scan_iname, sweep_iname, within_inames):
    domain = kernel.get_inames_domain(frozenset((sweep_iname, scan_iname)))
    domain = _move_set_to_param_dims_except(domain, (sweep_iname, scan_iname))

    var_dict = domain.get_var_dict()
    sweep_idx = var_dict[sweep_iname][1]
    scan_idx = var_dict[scan_iname][1]

    domain = domain.project_out_except(
            within_inames | kernel.non_iname_variable_names(), (isl.dim_type.param,))

    try:
        sweep_lower_bound = domain.dim_min(sweep_idx)
        sweep_upper_bound = domain.dim_max(sweep_idx)
        scan_lower_bound = domain.dim_min(scan_idx)
    except isl.Error as e:
        raise ValueError("isl error: %s" % e) from e

    return (sweep_lower_bound, sweep_upper_bound, scan_lower_bound)


def _try_infer_scan_stride(kernel, scan_iname, sweep_iname, sweep_lower_bound):
    """The stride is the number of steps the scan iname takes per iteration
    of the sweep iname. This is allowed to be an integer constant.

    E.g. for a domain of {[i,j]: 0<=i<n and 0<=j<=6*i}, the stride is 6.
    """
    dim_type = isl.dim_type

    domain = kernel.get_inames_domain(frozenset([sweep_iname, scan_iname]))
    domain_with_sweep_param = _move_set_to_param_dims_except(domain, (scan_iname,))

    domain_with_sweep_param = domain_with_sweep_param.project_out_except(
            (sweep_iname, scan_iname), (dim_type.set, dim_type.param))

    scan_iname_idx = domain_with_sweep_param.find_dim_by_name(
            dim_type.set, scan_iname)

    # Should be equal to k * sweep_iname, where k is the stride.

    try:
        scan_iname_range = (
                domain_with_sweep_param.dim_max(scan_iname_idx)
                - domain_with_sweep_param.dim_min(scan_iname_idx)
                ).gist(domain_with_sweep_param.params())
    except isl.Error as e:
        raise ValueError("isl error: '%s'" % e) from e

    scan_iname_pieces = scan_iname_range.get_pieces()

    if len(scan_iname_pieces) > 1:
        raise ValueError("range in multiple pieces: %s" % scan_iname_range)
    elif len(scan_iname_pieces) == 0:
        raise ValueError("empty range found for iname '%s'" % scan_iname)

    scan_iname_constr, scan_iname_aff = scan_iname_pieces[0]

    if not scan_iname_constr.plain_is_universe():
        raise ValueError("found constraints: %s" % scan_iname_constr)

    if scan_iname_aff.dim(dim_type.div):
        raise ValueError("aff has div: %s" % scan_iname_aff)

    coeffs = scan_iname_aff.get_coefficients_by_name(dim_type.param)

    if len(coeffs) == 0:
        try:
            scan_iname_aff.get_constant_val()
        except Exception as err:
            raise ValueError(
                        "range for aff isn't constant: '%s'" % scan_iname_aff) from err

        # If this point is reached we're assuming the domain is of the form
        # {[i,j]: i=0 and j=0}, so the stride is technically 1 - any value
        # this function returns will be verified later by
        # _check_reduction_is_triangular().
        return 1

    if sweep_iname not in coeffs:
        raise ValueError("didn't find sweep iname in coeffs: %s" % sweep_iname)

    stride = coeffs[sweep_iname]

    if not stride.is_int():
        raise ValueError("stride not an integer: %s" % stride)

    if not stride.is_pos():
        raise ValueError("stride not positive: %s" % stride)

    return stride.to_python()

# }}}


# {{{ domain creation for scans

def _get_domain_with_iname_as_param(domain, iname):
    dim_type = isl.dim_type

    if domain.find_dim_by_name(dim_type.param, iname) >= 0:
        return domain

    iname_idx = domain.find_dim_by_name(dim_type.set, iname)

    assert iname_idx >= 0, (iname, domain)

    return domain.move_dims(
        dim_type.param, domain.dim(dim_type.param),
        dim_type.set, iname_idx, 1)


def _create_domain_for_sweep_tracking(orig_domain, tracking_iname, scan_param):
    sp = scan_param

    dim_type = isl.dim_type

    subd = isl.BasicSet.universe(orig_domain.params().space)

    # Add tracking_iname and sweep iname.

    subd = _add_params_to_domain(subd, (sp.sweep_iname, tracking_iname))

    # Here we realize the domain:
    #
    # [..., i] -> {
    #  [j]: 0 <= j - l
    #       and
    #       j - l <= k * (i - m)
    #       and
    #       k * (i - m - 1) < j - l }
    # where
    #   * i is the sweep iname
    #   * j is the tracking iname
    #   * k is the stride for the scan
    #   * l is the lower bound for the scan
    #   * m is the lower bound for the sweep iname
    #
    affs = isl.affs_from_space(subd.space)

    subd &= (affs[tracking_iname] - sp.scan_lower_bound).ge_set(affs[0])
    subd &= (affs[tracking_iname] - sp.scan_lower_bound)\
            .le_set(sp.stride * (affs[sp.sweep_iname] - sp.sweep_lower_bound))
    subd &= (affs[tracking_iname] - sp.scan_lower_bound)\
            .gt_set(sp.stride * (affs[sp.sweep_iname] - sp.sweep_lower_bound - 1))

    # Move tracking_iname into a set dim (NOT sweep iname).
    subd = subd.move_dims(
            dim_type.set, 0,
            dim_type.param, subd.dim(dim_type.param) - 1, 1)

    # Simplify (maybe).
    orig_domain_with_sweep_param = (
            _get_domain_with_iname_as_param(orig_domain, sp.sweep_iname))
    subd = subd.gist_params(orig_domain_with_sweep_param.params())

    subd, = subd.get_basic_sets()

    return subd

# }}}


# {{{ _hackily_ensure_multi_assignment_return_values_are_scoped_private

def _hackily_ensure_multi_assignment_return_values_are_scoped_private(kernel):
    """
    Multi assignment function calls are currently lowered into OpenCL so that
    the function call::

       a, b = segmented_sum(x, y, z, w)

    becomes::

       a = segmented_sum_mangled(x, y, z, w, &b).

    For OpenCL, the scope of "b" is significant, and the preamble generation
    currently assumes the scope is always private. This function forces that to
    be the case by introducing temporary assignments into the kernel.
    """

    insn_id_gen = kernel.get_instruction_id_generator()
    var_name_gen = kernel.get_var_name_generator()

    new_or_updated_instructions = {}
    new_temporaries = {}

    dep_map = {
            insn.id: insn.depends_on for insn in kernel.instructions}

    inverse_dep_map = {insn.id: set() for insn in kernel.instructions}

    for insn_id, deps in dep_map.items():
        for dep in deps:
            inverse_dep_map[dep].add(insn_id)

    del dep_map

    # {{{ utils

    def _add_to_no_sync_with(insn_id, new_no_sync_with_params):
        insn = kernel.id_to_insn.get(insn_id)
        insn = new_or_updated_instructions.get(insn_id, insn)
        new_or_updated_instructions[insn_id] = (
                insn.copy(
                    no_sync_with=(
                        insn.no_sync_with | frozenset(new_no_sync_with_params))))

    def _add_to_depends_on(insn_id, new_depends_on_params):
        insn = kernel.id_to_insn.get(insn_id)
        insn = new_or_updated_instructions.get(insn_id, insn)
        new_or_updated_instructions[insn_id] = (
                insn.copy(
                    depends_on=insn.depends_on | frozenset(new_depends_on_params)))

    # }}}

    from loopy.kernel.instruction import CallInstruction, is_array_call
    for insn in kernel.instructions:
        if not isinstance(insn, CallInstruction):
            continue

        if len(insn.assignees) <= 1:
            continue

        if is_array_call(insn.assignees, insn.expression):
            continue

        assignees = insn.assignees
        assignee_var_names = insn.assignee_var_names()

        new_assignees = [assignees[0]]
        newly_added_assignments_ids = set()
        needs_replacement = False

        last_added_insn_id = insn.id

        FIRST_POINTER_ASSIGNEE_IDX = 1  # noqa

        for assignee_nr, assignee_var_name, assignee in zip(
                range(FIRST_POINTER_ASSIGNEE_IDX, len(assignees)),
                assignee_var_names[FIRST_POINTER_ASSIGNEE_IDX:],
                assignees[FIRST_POINTER_ASSIGNEE_IDX:]):

            if (
                    assignee_var_name in kernel.temporary_variables
                    and
                    (kernel.temporary_variables[assignee_var_name].address_space
                         == AddressSpace.PRIVATE)):
                new_assignees.append(assignee)
                continue

            needs_replacement = True

            # {{{ generate a new assignment instruction

            new_assignee_name = var_name_gen(
                    "{insn_id}_retval_{assignee_nr}"
                    .format(insn_id=insn.id, assignee_nr=assignee_nr))

            new_assignment_id = insn_id_gen(
                    "{insn_id}_assign_retval_{assignee_nr}"
                    .format(insn_id=insn.id, assignee_nr=assignee_nr))

            newly_added_assignments_ids.add(new_assignment_id)

            new_temporaries[new_assignee_name] = (
                    TemporaryVariable(
                        name=new_assignee_name,
                        dtype=None,
                        shape=(),
                        address_space=AddressSpace.PRIVATE))

            from pymbolic import var
            new_assignee = var(new_assignee_name)
            new_assignees.append(new_assignee)

            new_or_updated_instructions[new_assignment_id] = (
                    make_assignment(
                        assignees=(assignee,),
                        expression=new_assignee,
                        id=new_assignment_id,
                        depends_on=frozenset([last_added_insn_id]),
                        depends_on_is_final=True,
                        no_sync_with=(
                            insn.no_sync_with | frozenset([(insn.id, "any")])),
                        predicates=insn.predicates,
                        within_inames=insn.within_inames))

            last_added_insn_id = new_assignment_id

            # }}}

        if not needs_replacement:
            continue

        # {{{ update originating instruction

        orig_insn = new_or_updated_instructions.get(insn.id, insn)

        new_or_updated_instructions[insn.id] = (
                orig_insn.copy(assignees=tuple(new_assignees)))

        _add_to_no_sync_with(insn.id,
                [(id, "any") for id in newly_added_assignments_ids])

        # }}}

        # {{{ squash spurious memory dependencies amongst new assignments

        for new_insn_id in newly_added_assignments_ids:
            _add_to_no_sync_with(new_insn_id,
                    [(id, "any")
                     for id in newly_added_assignments_ids
                     if id != new_insn_id])

        # }}}

        # {{{ update instructions that depend on the originating instruction

        for inverse_dep in inverse_dep_map[insn.id]:
            _add_to_depends_on(inverse_dep, newly_added_assignments_ids)

            for insn_id, scope in (
                    new_or_updated_instructions[inverse_dep].no_sync_with):
                if insn_id == insn.id:
                    _add_to_no_sync_with(
                            inverse_dep,
                            [(id, scope) for id in newly_added_assignments_ids])

        # }}}

    if not new_temporaries and not new_or_updated_instructions:
        return kernel

    new_temporary_variables = kernel.temporary_variables.copy()
    new_temporary_variables.update(new_temporaries)

    new_instructions = (
            list(new_or_updated_instructions.values())
            + [insn
                for insn in kernel.instructions
                if insn.id not in new_or_updated_instructions])

    return kernel.copy(temporary_variables=new_temporary_variables,
                       instructions=new_instructions)

# }}}


# {{{ RealizeReductionCallbackMapper

class RealizeReductionCallbackMapper(ReductionCallbackMapper):
    def __init__(self, callback):
        super().__init__(callback)

    def map_reduction(self, expr, **kwargs):
        return self.callback(expr, **kwargs)

    def map_if(self, expr, *, red_realize_ctx, nresults):
        common_kwargs = {"nresults": nresults}

        # {{{ generate code for condition

        rrc_cond = replace(red_realize_ctx,
                surrounding_insn_add_depends_on=set(),
                surrounding_insn_add_no_sync_with=set(),
                surrounding_insn_add_within_inames=set())
        import pymbolic.primitives as prim
        rec_cond = self.rec(
                expr.condition,
                red_realize_ctx=rrc_cond,
                **common_kwargs)
        assert not rrc_cond.surrounding_insn_add_no_sync_with
        assert not rrc_cond.surrounding_insn_add_within_inames

        cond_dep_on = rrc_cond.surrounding_insn_add_depends_on
        red_realize_ctx.surrounding_insn_add_depends_on.update(cond_dep_on)

        # }}}

        return prim.If(rec_cond,
                       self.rec(expr.then,
                           red_realize_ctx=replace(
                               red_realize_ctx,
                               surrounding_depends_on=(
                                   red_realize_ctx.surrounding_depends_on
                                   | cond_dep_on),
                               surrounding_predicates=(
                                   red_realize_ctx.surrounding_predicates
                                   | frozenset([rec_cond])
                                   )),
                           **common_kwargs),
                       self.rec(expr.else_,
                           red_realize_ctx=replace(
                               red_realize_ctx,
                               surrounding_depends_on=(
                                   red_realize_ctx.surrounding_depends_on
                                   | cond_dep_on),
                               surrounding_predicates=(
                                   red_realize_ctx.surrounding_predicates
                                   | frozenset([prim.LogicalNot(rec_cond)])
                                   )),
                           **common_kwargs))

# }}}


# {{{ helpers

def _strip_if_scalar(reference, val):
    if len(reference) == 1:
        return val[0]
    else:
        return val


def _preprocess_scan_arguments(
        red_realize_ctx,
        expr, nresults, scan_iname, track_iname,
        newly_generated_insn_id_set):
    """Does iname substitution within scan arguments and returns a set of values
    suitable to be passed to the binary op. Returns a tuple."""

    if nresults > 1:
        inner_expr = expr

        # In the case of a multi-argument scan, we need a name for each of
        # the arguments in order to pass them to the binary op - so we expand
        # items that are not "plain" tuples here.
        if not isinstance(inner_expr, tuple):
            get_args_insn_id = red_realize_ctx.insn_id_gen(
                    f"{red_realize_ctx.id_prefix}_{'_'.join(expr.inames)}_get")

            inner_expr = expand_inner_reduction(
                    red_realize_ctx=red_realize_ctx,
                    id=get_args_insn_id,
                    expr=inner_expr,
                    nresults=nresults,
                    depends_on=red_realize_ctx.surrounding_depends_on,
                    within_inames=red_realize_ctx.surrounding_within_inames,
                    predicates=red_realize_ctx.surrounding_predicates,
                    )

            newly_generated_insn_id_set = (
                    newly_generated_insn_id_set | frozenset({get_args_insn_id}))

        updated_inner_exprs = tuple(
                replace_var_within_expr(
                    red_realize_ctx.kernel, red_realize_ctx.var_name_gen,
                    sub_expr, scan_iname, track_iname)
                for sub_expr in inner_expr)
    else:
        updated_inner_exprs = (
                replace_var_within_expr(
                    red_realize_ctx.kernel, red_realize_ctx.var_name_gen,
                    expr, scan_iname, track_iname),)

    return updated_inner_exprs, newly_generated_insn_id_set

# }}}


def expand_inner_reduction(
        red_realize_ctx, id, expr, nresults, depends_on, within_inames, predicates):
    # FIXME: use _make_temporaries
    from pymbolic.primitives import Call

    from loopy.symbolic import Reduction
    assert isinstance(expr, (Call, Reduction))

    temp_var_names = [
            red_realize_ctx.var_name_gen(id + "_arg" + str(i))
            for i in range(nresults)]

    for name in temp_var_names:
        red_realize_ctx.additional_temporary_variables[name] = TemporaryVariable(
                name=name,
                shape=(),
                dtype=None,
                address_space=AddressSpace.PRIVATE)

    from pymbolic import var
    temp_vars = tuple(var(n) for n in temp_var_names)

    call_insn = make_assignment(
            id=id,
            assignees=temp_vars,
            expression=expr,
            depends_on=depends_on,
            within_inames=within_inames,
            within_inames_is_final=True,
            predicates=predicates)

    red_realize_ctx.additional_insns.append(call_insn)

    return temp_vars


# {{{ reduction type: sequential

def map_reduction_seq(red_realize_ctx, expr, nresults, arg_dtypes, reduction_dtypes):
    orig_kernel = red_realize_ctx.orig_kernel

    acc_var_names = _make_temporaries(
            red_realize_ctx=red_realize_ctx,
            name_based_on="acc_"+"_".join(expr.inames),
            nvars=nresults,
            shape=(),
            dtypes=reduction_dtypes,
            address_space=AddressSpace.PRIVATE)

    from pymbolic import var
    acc_vars = tuple(var(n) for n in acc_var_names)

    init_id = red_realize_ctx.insn_id_gen(
            f"{red_realize_ctx.id_prefix}_{'_'.join(expr.inames)}_init")

    expression, red_realize_ctx.boxed_callables_table[0] = \
            expr.operation.neutral_element(
                    *arg_dtypes,
                    callables_table=red_realize_ctx.boxed_callables_table[0],
                    target=red_realize_ctx.orig_kernel.target)

    init_insn = make_assignment(
            id=init_id,
            assignees=acc_vars,
            within_inames=red_realize_ctx.surrounding_within_inames,
            within_inames_is_final=True,
            depends_on=frozenset(),
            expression=expression,

            # Do not inherit predicates: Those might read variables
            # that may not yet be set, and we don't have a great way
            # of figuring out what the dependencies of the accumulator
            # initializer should be.

            # This way, we may initialize a few too many accumulators,
            # but that's better than being incorrect.
            # https://github.com/inducer/loopy/issues/231
            )

    red_realize_ctx.additional_insns.append(init_insn)

    update_id = red_realize_ctx.insn_id_gen(
            based_on=f"{red_realize_ctx.id_prefix}_{'_'.join(expr.inames)}_update")

    update_red_realize_ctx = red_realize_ctx.new_subinstruction(
            within_inames=(
                red_realize_ctx.surrounding_within_inames
                | frozenset(expr.inames)),
            depends_on=(
                frozenset({init_id})
                | red_realize_ctx.surrounding_depends_on))

    reduction_expr = red_realize_ctx.mapper(
            expr.expr, red_realize_ctx=update_red_realize_ctx,
            nresults=1)

    # In the case of a multi-argument reduction, we need a name for each of
    # the arguments in order to pass them to the binary op - so we expand
    # items that are not "plain" tuples here.
    if nresults > 1 and not isinstance(reduction_expr, tuple):
        get_args_insn_id = red_realize_ctx.insn_id_gen(
                f"{red_realize_ctx.id_prefix}_{'_'.join(expr.inames)}_get")

        reduction_expr = expand_inner_reduction(
                red_realize_ctx=red_realize_ctx,
                id=get_args_insn_id,
                expr=reduction_expr,
                nresults=nresults,
                depends_on=red_realize_ctx.surrounding_depends_on,
                within_inames=update_red_realize_ctx.surrounding_within_inames,
                predicates=red_realize_ctx.surrounding_predicates,
                )

        update_red_realize_ctx.surrounding_insn_add_depends_on.add(get_args_insn_id)

    expression, red_realize_ctx.boxed_callables_table[0] = expr.operation(
            arg_dtypes,
            _strip_if_scalar(acc_vars, acc_vars),
            reduction_expr,
            red_realize_ctx.boxed_callables_table[0],
            orig_kernel.target)

    reduction_insn = make_assignment(
            id=update_id,
            assignees=acc_vars,
            expression=expression,
            **update_red_realize_ctx.get_insn_kwargs())

    red_realize_ctx.additional_insns.append(reduction_insn)
    red_realize_ctx.surrounding_insn_add_depends_on.add(reduction_insn.id)

    if nresults == 1:
        assert len(acc_vars) == 1
        return acc_vars[0]
    else:
        return acc_vars

# }}}


# {{{ reduction type: local-parallel

def _get_int_iname_size(kernel, iname):
    from loopy.isl_helpers import static_max_of_pw_aff
    from loopy.symbolic import pw_aff_to_expr
    size = pw_aff_to_expr(
            static_max_of_pw_aff(
                kernel.get_iname_bounds(iname).size,
                constants_only=True))
    assert isinstance(size, int)
    return size


def _make_slab_set(iname, size):
    v = isl.make_zero_and_vars([iname])
    bs, = (
            v[0].le_set(v[iname])
            &
            v[iname].lt_set(v[0] + size)).get_basic_sets()
    return bs


def _make_slab_set_from_range(iname, lbound, ubound):
    v = isl.make_zero_and_vars([iname])
    bs, = (
            v[iname].ge_set(v[0] + lbound)
            &
            v[iname].lt_set(v[0] + ubound)).get_basic_sets()
    return bs


def map_reduction_local(red_realize_ctx, expr, nresults, arg_dtypes,
        reduction_dtypes):
    orig_kernel = red_realize_ctx.orig_kernel

    red_iname, = expr.inames

    size = _get_int_iname_size(orig_kernel, red_iname)

    from loopy.kernel.data import LocalInameTagBase
    surrounding_local_inames = tuple(
            oiname for oiname in red_realize_ctx.surrounding_within_inames
            if orig_kernel.iname_tags_of_type(oiname, LocalInameTagBase))

    from pymbolic import var
    outer_local_iname_vars = tuple(
            var(oiname) for oiname in surrounding_local_inames)

    outer_local_iname_sizes = tuple(
            _get_int_iname_size(orig_kernel, oiname)
            for oiname in surrounding_local_inames)

    neutral_var_names = _make_temporaries(
            red_realize_ctx=red_realize_ctx,
            name_based_on="neutral_"+red_iname,
            nvars=nresults,
            shape=(),
            dtypes=reduction_dtypes,
            address_space=AddressSpace.PRIVATE)

    acc_var_names = _make_temporaries(
            red_realize_ctx=red_realize_ctx,
            name_based_on="acc_"+red_iname,
            nvars=nresults,
            shape=(*outer_local_iname_sizes, size),
            dtypes=reduction_dtypes,
            address_space=AddressSpace.LOCAL)

    acc_vars = tuple(var(n) for n in acc_var_names)

    # {{{ add separate iname to carry out the reduction

    # Doing this sheds any odd conditionals that may be active
    # on our red_iname.

    base_exec_iname = red_realize_ctx.var_name_gen("red_"+red_iname)
    red_realize_ctx.domains.append(_make_slab_set(base_exec_iname, size))
    red_realize_ctx.additional_iname_tags[base_exec_iname] \
            = orig_kernel.iname_tags(red_iname)

    # }}}

    neutral, red_realize_ctx.boxed_callables_table[0] = \
            expr.operation.neutral_element(*arg_dtypes,
                    callables_table=red_realize_ctx.boxed_callables_table[0],
                    target=orig_kernel.target)
    init_id = red_realize_ctx.insn_id_gen(
            f"{red_realize_ctx.id_prefix}_{red_iname}_init")
    init_insn = make_assignment(
            id=init_id,
            assignees=tuple(
                acc_var[(*outer_local_iname_vars, var(base_exec_iname))]
                for acc_var in acc_vars),
            expression=neutral,
            within_inames=(
                red_realize_ctx.surrounding_within_inames
                | frozenset([base_exec_iname])),
            within_inames_is_final=True,
            depends_on=frozenset(),
            # Do not inherit predicates: Those might read variables
            # that may not yet be set, and we don't have a great way
            # of figuring out what the dependencies of the accumulator
            # initializer should be.

            # This way, we may initialize a few too many accumulators,
            # but that's better than being incorrect.
            # https://github.com/inducer/loopy/issues/231
            )
    red_realize_ctx.additional_insns.append(init_insn)

    init_neutral_id = red_realize_ctx.insn_id_gen(
            f"{red_realize_ctx.id_prefix}_{red_iname}_init_neutral")
    init_neutral_insn = make_assignment(
            id=init_neutral_id,
            assignees=tuple(var(nvn) for nvn in neutral_var_names),
            expression=neutral,
            within_inames=(
                red_realize_ctx.surrounding_within_inames
                | frozenset([base_exec_iname])),
            within_inames_is_final=True,
            depends_on=frozenset(),
            predicates=red_realize_ctx.surrounding_predicates,
            )
    red_realize_ctx.additional_insns.append(init_neutral_insn)

    transfer_depends_on = {init_neutral_id, init_id}

    transfer_red_realize_ctx = red_realize_ctx.new_subinstruction(
            within_inames=(
                    red_realize_ctx.surrounding_within_inames
                    | frozenset([red_iname])),
            depends_on=(
                red_realize_ctx.surrounding_depends_on
                | frozenset([init_id, init_neutral_id])),
            no_sync_with=(
                red_realize_ctx.surrounding_no_sync_with
                | frozenset([(init_id, "any")])))

    reduction_expr = red_realize_ctx.mapper(
            expr.expr, red_realize_ctx=transfer_red_realize_ctx,
            nresults=1)

    # In the case of a multi-argument reduction, we need a name for each of
    # the arguments in order to pass them to the binary op - so we expand
    # items that are not "plain" tuples here.
    if nresults > 1 and not isinstance(reduction_expr, tuple):
        get_args_insn_id = red_realize_ctx.insn_id_gen(
                f"{red_realize_ctx.id_prefix}_{red_iname}_get")

        reduction_expr = expand_inner_reduction(
                red_realize_ctx=red_realize_ctx,
                id=get_args_insn_id,
                expr=reduction_expr,
                nresults=nresults,
                depends_on=red_realize_ctx.surrounding_depends_on,
                within_inames=transfer_red_realize_ctx.surrounding_within_inames,
                predicates=red_realize_ctx.surrounding_predicates,
                )

        transfer_depends_on.add(get_args_insn_id)

    transfer_id = red_realize_ctx.insn_id_gen(
            f"{red_realize_ctx.id_prefix}_{red_iname}_transfer")
    expression, red_realize_ctx.boxed_callables_table[0] = expr.operation(
            arg_dtypes,
            _strip_if_scalar(
                neutral_var_names,
                tuple(var(nvn) for nvn in neutral_var_names)),
            reduction_expr,
            red_realize_ctx.boxed_callables_table[0],
            orig_kernel.target)
    transfer_insn = make_assignment(
            id=transfer_id,
            assignees=tuple(
                acc_var[(*outer_local_iname_vars, var(red_iname))]
                for acc_var in acc_vars),
            expression=expression,
            **transfer_red_realize_ctx.get_insn_kwargs())
    red_realize_ctx.additional_insns.append(transfer_insn)

    cur_size = 1
    while cur_size < size:
        cur_size *= 2

    prev_id = transfer_id
    bound = size

    stage_exec_iname = None

    istage = 0
    while cur_size > 1:

        new_size = cur_size // 2
        assert new_size * 2 == cur_size

        stage_exec_iname = red_realize_ctx.var_name_gen(
                "red_%s_s%d" % (red_iname, istage))
        red_realize_ctx.domains.append(
                _make_slab_set(stage_exec_iname, bound-new_size))
        red_realize_ctx.additional_iname_tags[stage_exec_iname] \
                = orig_kernel.iname_tags(red_iname)

        stage_id = red_realize_ctx.insn_id_gen(
                "red_%s_stage_%d" % (red_iname, istage))

        expression, red_realize_ctx.boxed_callables_table[0] = expr.operation(
                arg_dtypes,
                _strip_if_scalar(acc_vars, tuple(
                    acc_var[
                        (*outer_local_iname_vars, var(stage_exec_iname))]
                    for acc_var in acc_vars)),
                _strip_if_scalar(acc_vars, tuple(
                    acc_var[
                        (*outer_local_iname_vars, var(stage_exec_iname) + new_size)]
                    for acc_var in acc_vars)),
                red_realize_ctx.boxed_callables_table[0],
                orig_kernel.target)

        stage_insn = make_assignment(
                id=stage_id,
                assignees=tuple(
                    acc_var[(*outer_local_iname_vars, var(stage_exec_iname))]
                    for acc_var in acc_vars),
                expression=expression,
                within_inames=(
                    red_realize_ctx.surrounding_within_inames
                    | frozenset([stage_exec_iname])),
                within_inames_is_final=True,
                depends_on=frozenset([prev_id]),
                predicates=red_realize_ctx.surrounding_predicates,
                )

        red_realize_ctx.additional_insns.append(stage_insn)
        prev_id = stage_id

        cur_size = new_size
        bound = cur_size
        istage += 1

    red_realize_ctx.surrounding_insn_add_depends_on.add(prev_id)
    red_realize_ctx.surrounding_insn_add_no_sync_with.add((prev_id, "any"))
    red_realize_ctx.surrounding_insn_add_within_inames.add(
            stage_exec_iname or base_exec_iname)

    if nresults == 1:
        assert len(acc_vars) == 1
        return acc_vars[0][(*outer_local_iname_vars, 0)]
    else:
        return [acc_var[(*outer_local_iname_vars, 0)] for acc_var in
                acc_vars]
# }}}


# {{{ utils (stateful)

@memoize_on_first_arg
def _get_or_add_sweep_tracking_iname_and_domain(
        red_realize_ctx,
        scan_param,
        tracking_iname):
    kernel = red_realize_ctx.kernel

    domain = kernel.get_inames_domain(
            frozenset((scan_param.scan_iname, scan_param.sweep_iname)))

    red_realize_ctx.inames_added_for_scan.add(tracking_iname)

    new_domain = _create_domain_for_sweep_tracking(
            domain, tracking_iname, scan_param)

    _insert_subdomain_into_domain_tree(kernel, red_realize_ctx.domains, new_domain)

    return tracking_iname


def replace_var_within_expr(kernel, var_name_gen, expr, from_var, to_var):
    from pymbolic.mapper.substitutor import make_subst_func

    from loopy.symbolic import (
        RuleAwareSubstitutionMapper,
        SubstitutionRuleMappingContext,
    )

    # FIXME: This is broken. SubstitutionRuleMappingContext produces a new
    # kernel (via finish_kernel) with new subst rules. These get dropped on the
    # floor here.
    rule_mapping_context = SubstitutionRuleMappingContext(
        kernel.substitutions, var_name_gen)

    from pymbolic import var
    mapper = RuleAwareSubstitutionMapper(
        rule_mapping_context,
        make_subst_func({from_var: var(to_var)}),
        within=lambda *args: True)

    return mapper(expr, kernel, None)


def _make_temporaries(
        red_realize_ctx, name_based_on, nvars, shape, dtypes, address_space):
    var_names = [
            red_realize_ctx.var_name_gen(name_based_on.format(index=i))
            for i in range(nvars)]

    from loopy.kernel.data import TemporaryVariable

    for name, dtype in zip(var_names, dtypes):
        red_realize_ctx.additional_temporary_variables[name] = TemporaryVariable(
                name=name,
                shape=shape,
                dtype=dtype,
                address_space=address_space)

    return var_names

# }}}


# {{{ reduction type: sequential scan

def map_scan_seq(red_realize_ctx, expr, nresults, arg_dtypes,
        reduction_dtypes, scan_param):

    track_iname = red_realize_ctx.var_name_gen(
            "{sweep_iname}__seq_scan"
            .format(sweep_iname=scan_param.sweep_iname))

    _get_or_add_sweep_tracking_iname_and_domain(
            red_realize_ctx, scan_param, track_iname)
    red_realize_ctx.additional_iname_tags[track_iname] = frozenset()

    from loopy.kernel.data import AddressSpace
    acc_var_names = _make_temporaries(
            red_realize_ctx=red_realize_ctx,
            name_based_on="acc_" + scan_param.scan_iname,
            nvars=nresults,
            shape=(),
            dtypes=reduction_dtypes,
            address_space=AddressSpace.PRIVATE)

    from pymbolic import var
    acc_vars = tuple(var(n) for n in acc_var_names)

    init_id = red_realize_ctx.insn_id_gen(
            f"{red_realize_ctx.id_prefix}_{'_'.join(expr.inames)}_init")

    init_insn_depends_on = frozenset()

    expression, red_realize_ctx.boxed_callables_table[0] = \
            expr.operation.neutral_element(*arg_dtypes,
                    callables_table=red_realize_ctx.boxed_callables_table[0],
                    target=red_realize_ctx.orig_kernel.target)

    init_insn = make_assignment(
            id=init_id,
            assignees=acc_vars,
            within_inames=(
                red_realize_ctx.surrounding_within_inames
                - frozenset((scan_param.sweep_iname, *expr.inames))),
            within_inames_is_final=True,
            depends_on=init_insn_depends_on,
            expression=expression,
            # Do not inherit predicates: Those might read variables
            # that may not yet be set, and we don't have a great way
            # of figuring out what the dependencies of the accumulator
            # initializer should be.

            # This way, we may initialize a few too many accumulators,
            # but that's better than being incorrect.
            # https://github.com/inducer/loopy/issues/231
            )

    red_realize_ctx.additional_insns.append(init_insn)

    scan_insn_depends_on = {init_insn.id} | red_realize_ctx.surrounding_depends_on

    scan_red_realize_ctx = red_realize_ctx.new_subinstruction(
            within_inames=(
                red_realize_ctx.surrounding_within_inames
                | frozenset({scan_param.scan_iname})),
            depends_on=red_realize_ctx.surrounding_depends_on)

    reduction_expr = red_realize_ctx.mapper(
            expr.expr, red_realize_ctx=scan_red_realize_ctx,
            nresults=1)

    updated_inner_exprs, scan_insn_depends_on = _preprocess_scan_arguments(
            scan_red_realize_ctx,
            reduction_expr, nresults,
            scan_param.scan_iname, track_iname, scan_insn_depends_on)

    scan_id = red_realize_ctx.insn_id_gen(
            based_on=f"{red_realize_ctx.id_prefix}_{'_'.join(expr.inames)}_scan")

    expression, red_realize_ctx.boxed_callables_table[0] = expr.operation(
            arg_dtypes,
            _strip_if_scalar(acc_vars, acc_vars),
            _strip_if_scalar(acc_vars, updated_inner_exprs),
            red_realize_ctx.boxed_callables_table[0],
            red_realize_ctx.orig_kernel.target)

    scan_insn = make_assignment(
            id=scan_id,
            assignees=acc_vars,
            expression=expression,
            within_inames=(
                red_realize_ctx.surrounding_within_inames
                | frozenset(
                    scan_red_realize_ctx.surrounding_insn_add_within_inames)
                | {track_iname}),
            depends_on=(
                frozenset(scan_insn_depends_on)
                | frozenset(scan_red_realize_ctx.surrounding_insn_add_depends_on)
                ),
            no_sync_with=(
                red_realize_ctx.surrounding_no_sync_with
                | frozenset(scan_red_realize_ctx.surrounding_insn_add_no_sync_with)
                ),
            within_inames_is_final=True,
            predicates=red_realize_ctx.surrounding_predicates,
            )

    red_realize_ctx.additional_insns.append(scan_insn)
    red_realize_ctx.surrounding_insn_add_depends_on.add(scan_insn.id)

    if nresults == 1:
        assert len(acc_vars) == 1
        return acc_vars[0]
    else:
        return acc_vars

# }}}


# {{{ reduction type: local-parallel scan

def map_scan_local(red_realize_ctx, expr, nresults, arg_dtypes,
        reduction_dtypes, scan_param):

    orig_kernel = red_realize_ctx.orig_kernel

    scan_size = _get_int_iname_size(orig_kernel, scan_param.sweep_iname)

    assert scan_size > 0

    if scan_size == 1:
        return map_reduction_seq(red_realize_ctx,
                expr, nresults, arg_dtypes, reduction_dtypes)

    from loopy.kernel.data import LocalInameTagBase
    outer_local_inames = tuple(
            oiname for oiname in red_realize_ctx.surrounding_within_inames
            if orig_kernel.iname_tags_of_type(oiname, LocalInameTagBase)
            and oiname != scan_param.sweep_iname)

    from pymbolic import var
    outer_local_iname_vars = tuple(
            var(oiname) for oiname in outer_local_inames)

    outer_local_iname_sizes = tuple(
            _get_int_iname_size(orig_kernel, oiname)
            for oiname in outer_local_inames)

    track_iname = red_realize_ctx.var_name_gen(
            "{sweep_iname}__pre_scan"
            .format(sweep_iname=scan_param.sweep_iname))

    _get_or_add_sweep_tracking_iname_and_domain(
            red_realize_ctx,
            scan_param,
            track_iname)
    red_realize_ctx.additional_iname_tags[track_iname] = frozenset()

    # {{{ add separate iname to carry out the scan

    # Doing this sheds any odd conditionals that may be active
    # on our scan_iname.

    base_exec_iname = red_realize_ctx.var_name_gen(scan_param.sweep_iname + "__scan")
    red_realize_ctx.domains.append(_make_slab_set(base_exec_iname, scan_size))
    red_realize_ctx.additional_iname_tags[base_exec_iname] \
            = orig_kernel.iname_tags(scan_param.sweep_iname)

    # }}}

    read_var_names = _make_temporaries(
            red_realize_ctx=red_realize_ctx,
            name_based_on="read_"+scan_param.scan_iname+"_arg_{index}",
            nvars=nresults,
            shape=(),
            dtypes=reduction_dtypes,
            address_space=AddressSpace.PRIVATE)

    acc_var_names = _make_temporaries(
            red_realize_ctx=red_realize_ctx,
            name_based_on="acc_"+scan_param.scan_iname,
            nvars=nresults,
            shape=(*outer_local_iname_sizes, scan_size),
            dtypes=reduction_dtypes,
            address_space=AddressSpace.LOCAL)

    acc_vars = tuple(var(n) for n in acc_var_names)
    read_vars = tuple(var(n) for n in read_var_names)

    base_iname_deps = (
            red_realize_ctx.surrounding_within_inames
            - frozenset([scan_param.sweep_iname]))

    neutral, red_realize_ctx.boxed_callables_table[0] = \
            expr.operation.neutral_element(*arg_dtypes,
                    callables_table=red_realize_ctx.boxed_callables_table[0],
                    target=orig_kernel.target)

    init_id = red_realize_ctx.insn_id_gen(
            f"{red_realize_ctx.id_prefix}_{scan_param.scan_iname}_init")
    init_insn = make_assignment(
            id=init_id,
            assignees=tuple(
                acc_var[(*outer_local_iname_vars, var(base_exec_iname))]
                for acc_var in acc_vars),
            expression=neutral,
            within_inames=base_iname_deps | frozenset([base_exec_iname]),
            within_inames_is_final=True,
            depends_on=frozenset(),
            # Do not inherit predicates: Those might read variables
            # that may not yet be set, and we don't have a great way
            # of figuring out what the dependencies of the accumulator
            # initializer should be.

            # This way, we may initialize a few too many accumulators,
            # but that's better than being incorrect.
            # https://github.com/inducer/loopy/issues/231
            )
    red_realize_ctx.additional_insns.append(init_insn)

    transfer_insn_depends_on = (
            frozenset({init_insn.id})
            | red_realize_ctx.surrounding_depends_on)

    transfer_red_realize_ctx = red_realize_ctx.new_subinstruction(
            within_inames=(
                red_realize_ctx.surrounding_within_inames
                | frozenset({scan_param.scan_iname})),
            depends_on=red_realize_ctx.surrounding_depends_on)

    reduction_expr = red_realize_ctx.mapper(
            expr.expr, red_realize_ctx=transfer_red_realize_ctx,
            nresults=1)

    updated_inner_exprs, transfer_insn_depends_on = _preprocess_scan_arguments(
            red_realize_ctx,
            reduction_expr, nresults,
            scan_param.scan_iname, track_iname, transfer_insn_depends_on)

    from loopy.symbolic import Reduction
    pre_scan_reduction = Reduction(
                operation=expr.operation,
                inames=(track_iname,),
                expr=_strip_if_scalar(acc_vars, updated_inner_exprs),
                allow_simultaneous=False,
                )

    pre_scan_result = red_realize_ctx.mapper(
            pre_scan_reduction, red_realize_ctx=transfer_red_realize_ctx,
            nresults=len(acc_vars))

    from loopy.symbolic import pw_aff_to_expr
    sweep_lower_bound_expr = pw_aff_to_expr(scan_param.sweep_lower_bound)

    if nresults == 1:
        assert not isinstance(pre_scan_result, tuple)
        pre_scan_result = (pre_scan_result,)

    transfer_ids = frozenset()
    for acc_var, pre_scan_result_i in zip(acc_vars, pre_scan_result):
        transfer_id = red_realize_ctx.insn_id_gen(
                f"{red_realize_ctx.id_prefix}_{scan_param.scan_iname}_transfer")
        transfer_insn = make_assignment(
                id=transfer_id,
                assignees=(acc_var[(
                    *outer_local_iname_vars,
                    var(scan_param.sweep_iname) - sweep_lower_bound_expr)
                ],),
                expression=pre_scan_result_i,
                within_inames=(
                    red_realize_ctx.surrounding_within_inames
                    | transfer_red_realize_ctx.surrounding_insn_add_within_inames
                    | frozenset({scan_param.sweep_iname})),
                within_inames_is_final=True,
                depends_on=(
                    transfer_insn_depends_on
                    | transfer_red_realize_ctx.surrounding_insn_add_depends_on),
                no_sync_with=(
                    frozenset([(init_id, "any")])
                    | transfer_red_realize_ctx.surrounding_insn_add_no_sync_with),
                predicates=red_realize_ctx.surrounding_predicates,
                )

        red_realize_ctx.additional_insns.append(transfer_insn)
        transfer_ids = transfer_ids | frozenset({transfer_id})

    del transfer_id

    prev_ids = transfer_ids

    istage = 0
    cur_size = 1

    while cur_size < scan_size:
        stage_exec_iname = red_realize_ctx.var_name_gen(
                f"{scan_param.sweep_iname}__scan_s{istage}")
        red_realize_ctx.domains.append(
                _make_slab_set_from_range(stage_exec_iname, cur_size, scan_size))
        red_realize_ctx.additional_iname_tags[stage_exec_iname] \
                = orig_kernel.iname_tags(scan_param.sweep_iname)

        for read_var, acc_var in zip(read_vars, acc_vars):
            read_stage_id = red_realize_ctx.insn_id_gen(
                    f"scan_{scan_param.scan_iname}_read_stage_{istage}")

            read_stage_insn = make_assignment(
                    id=read_stage_id,
                    assignees=(read_var,),
                    expression=(
                            acc_var[
                                (*outer_local_iname_vars,
                                    var(stage_exec_iname) - cur_size)]),
                    within_inames=(
                        base_iname_deps | frozenset([stage_exec_iname])),
                    within_inames_is_final=True,
                    depends_on=prev_ids,
                    predicates=red_realize_ctx.surrounding_predicates,
                    )

            if cur_size == 1:
                # Performance hack: don't add a barrier here with transfer_insn.
                # NOTE: This won't work if the way that local inames
                # are lowered changes.
                read_stage_insn = read_stage_insn.copy(
                        no_sync_with=(
                            read_stage_insn.no_sync_with
                            | frozenset([(tid, "any") for tid in transfer_ids])))

            red_realize_ctx.additional_insns.append(read_stage_insn)
            prev_ids = frozenset({read_stage_id})

        write_stage_id = red_realize_ctx.insn_id_gen(
                f"scan_{scan_param.scan_iname}_write_stage_{istage}")

        expression, red_realize_ctx.boxed_callables_table[0] = expr.operation(
            arg_dtypes,
            _strip_if_scalar(acc_vars, read_vars),
            _strip_if_scalar(acc_vars, tuple(
                acc_var[
                    (*outer_local_iname_vars, var(stage_exec_iname))]
                for acc_var in acc_vars)),
            red_realize_ctx.boxed_callables_table[0],
            orig_kernel.target)

        write_stage_insn = make_assignment(
                id=write_stage_id,
                assignees=tuple(
                    acc_var[(*outer_local_iname_vars, var(stage_exec_iname))]
                    for acc_var in acc_vars),
                expression=expression,
                within_inames=(
                    base_iname_deps | frozenset([stage_exec_iname])),
                within_inames_is_final=True,
                depends_on=prev_ids,
                predicates=red_realize_ctx.surrounding_predicates,
                )

        red_realize_ctx.additional_insns.append(write_stage_insn)
        prev_ids = frozenset({write_stage_id})

        cur_size *= 2
        istage += 1

    red_realize_ctx.surrounding_insn_add_depends_on.update(prev_ids)
    red_realize_ctx.surrounding_insn_add_within_inames.add(scan_param.sweep_iname)

    output_idx = var(scan_param.sweep_iname) - sweep_lower_bound_expr

    if nresults == 1:
        assert len(acc_vars) == 1
        return acc_vars[0][(*outer_local_iname_vars, output_idx)]
    else:
        return [acc_var[(*outer_local_iname_vars, output_idx)]
                for acc_var in acc_vars]

# }}}


# {{{ top-level dispatch among reduction types

def map_reduction(expr, *, red_realize_ctx, nresults):
    kernel_with_updated_domains = red_realize_ctx.kernel.copy(
            domains=red_realize_ctx.domains)

    from loopy.type_inference import (
        infer_arg_and_reduction_dtypes_for_reduction_expression,
    )
    arg_dtypes, reduction_dtypes = (
            infer_arg_and_reduction_dtypes_for_reduction_expression(
                kernel_with_updated_domains, expr,
                red_realize_ctx.boxed_callables_table[0],
                red_realize_ctx.unknown_types_ok))

    bad_inames = frozenset(expr.inames) & red_realize_ctx.surrounding_within_inames
    if bad_inames:
        raise LoopyError("reduction used within loop(s) that it was "
                "supposed to reduce over: " + ", ".join(bad_inames))

    iname_classes = _classify_reduction_inames(red_realize_ctx, expr.inames)

    n_sequential = len(iname_classes.sequential)
    n_local_par = len(iname_classes.local_parallel)
    n_nonlocal_par = len(iname_classes.nonlocal_parallel)

    really_force_scan = red_realize_ctx.force_scan and (
            len(expr.inames) != 1
            or expr.inames[0] not in red_realize_ctx.inames_added_for_scan)

    def _error_if_force_scan_on(cls, msg):
        if really_force_scan:
            raise cls(msg)

    may_be_implemented_as_scan = False
    if red_realize_ctx.force_scan or red_realize_ctx.automagic_scans_ok:
        try:
            # Try to determine scan candidate information (sweep iname, scan
            # iname, etc).
            scan_param = _try_infer_scan_candidate_from_expr(
                    kernel_with_updated_domains, expr,
                    red_realize_ctx.surrounding_within_inames,
                    sweep_iname=red_realize_ctx.force_outer_iname_for_scan)

        except ValueError as v:
            error = str(v)

        else:
            # Ensures the reduction is triangular (somewhat expensive).
            may_be_implemented_as_scan, error = _check_reduction_is_triangular(
                        kernel_with_updated_domains, expr, scan_param)

        if not may_be_implemented_as_scan:
            _error_if_force_scan_on(ReductionIsNotTriangularError, error)

    # {{{ sanity checks

    if n_local_par and n_sequential:
        raise LoopyError("Reduction over '%s' contains both parallel and "
                "sequential inames. It must be split "
                "(using split_reduction_{in,out}ward) "
                "before code generation."
                % ", ".join(expr.inames))

    if n_local_par > 1:
        raise LoopyError("Reduction over '%s' contains more than "
                "one parallel iname. It must be split "
                "(using split_reduction_{in,out}ward) "
                "before code generation."
                % ", ".join(expr.inames))

    if n_nonlocal_par:
        bad_inames = iname_classes.nonlocal_parallel
        raise LoopyError("the only form of parallelism supported "
                "by reductions is 'local'--found iname(s) '%s' "
                "respectively tagged '%s'"
                % (", ".join(bad_inames),
                   ", ".join(str(red_realize_ctx.orig_kernel.iname_tags(iname))
                             for iname in bad_inames)))

    # }}}

    red_realize_ctx.changes_made()

    if n_local_par == 0 and n_sequential == 0:
        warn_with_kernel(red_realize_ctx.kernel, "empty_reduction",
                "Empty reduction found (no inames to reduce over). "
                "Eliminating.")

        # We're not supposed to reduce/sum at all. (Note how this is distinct
        # from an empty reduction--there is an element here, just no inames
        # to reduce over. It's rather similar to an array with () shape in
        # numpy.)

        return expr.expr

    if may_be_implemented_as_scan:
        assert red_realize_ctx.force_scan or red_realize_ctx.automagic_scans_ok

        # We require the "scan" iname to be tagged sequential.
        if n_sequential:
            sweep_iname = scan_param.sweep_iname
            sweep_class = _classify_reduction_inames(red_realize_ctx, (sweep_iname,))

            sequential = sweep_iname in sweep_class.sequential
            parallel = sweep_iname in sweep_class.local_parallel
            bad_parallel = sweep_iname in sweep_class.nonlocal_parallel

            if sweep_iname not in red_realize_ctx.surrounding_within_inames:
                _error_if_force_scan_on(LoopyError,
                        "Sweep iname '%s' was detected, but is not an iname "
                        "for the instruction." % sweep_iname)
            elif bad_parallel:
                _error_if_force_scan_on(LoopyError,
                        "Sweep iname '%s' has an unsupported parallel tag '%s' "
                        "- the only parallelism allowed is 'local'." %
                        (sweep_iname,
                         ", ".join(tag.key
                        for tag in red_realize_ctx.kernel.iname_tags(sweep_iname))))
            elif parallel:
                return map_scan_local(red_realize_ctx, expr, nresults,
                        arg_dtypes, reduction_dtypes, scan_param)
            elif sequential:
                return map_scan_seq(red_realize_ctx, expr, nresults,
                        arg_dtypes, reduction_dtypes, scan_param)

            # fallthrough to reduction implementation

        else:
            assert n_local_par > 0
            scan_iname, = expr.inames
            _error_if_force_scan_on(LoopyError,
                    "Scan iname '%s' is parallel tagged: this is not allowed "
                    "(only the sweep iname should be tagged if parallelism "
                    "is desired)." % scan_iname)

            # fallthrough to reduction implementation

    if n_sequential:
        assert n_local_par == 0
        return map_reduction_seq(
                red_realize_ctx,
                expr, nresults, arg_dtypes, reduction_dtypes)
    else:
        assert n_local_par > 0
        return map_reduction_local(
                red_realize_ctx,
                expr, nresults, arg_dtypes,
                reduction_dtypes)

# }}}


# {{{ realize_reduction_for_single_kernel

# @remove_any_newly_unused_inames
def realize_reduction_for_single_kernel(kernel, callables_table,
        insn_id_filter=None, unknown_types_ok=True, automagic_scans_ok=False,
        force_scan=False, force_outer_iname_for_scan=None):
    logger.debug("%s: realize reduction" % kernel.name)

    orig_kernel = kernel

    finished_insns = []

    insn_id_gen = kernel.get_instruction_id_generator()
    var_name_gen = kernel.get_var_name_generator()

    cb_mapper = RealizeReductionCallbackMapper(map_reduction)

    insn_queue = kernel.instructions[:]
    domains = kernel.domains[:]

    inames_added_for_scan = set()

    kernel_changed = False

    if insn_id_filter is None:
        insn_id_filter = set(kernel.id_to_insn)
    elif isinstance(insn_id_filter, str):
        insn_id_filter = {insn_id_filter}
    else:
        from collections.abc import Collection
        if not isinstance(insn_id_filter, Collection):
            raise LoopyError("'insn_id_filter' can be either None, a string or a"
                             f" collection of strings. Got {type(insn_id_filter)}.")
        else:
            insn_id_filter = set(insn_id_filter)

    while insn_queue:
        insn = insn_queue.pop(0)

        red_realize_ctx = _ReductionRealizationContext(
                mapper=cb_mapper,

                force_scan=force_scan,
                automagic_scans_ok=automagic_scans_ok,
                unknown_types_ok=unknown_types_ok,
                force_outer_iname_for_scan=force_outer_iname_for_scan,

                orig_kernel=orig_kernel,
                kernel=kernel,

                id_prefix=insn.id,

                insn_id_gen=insn_id_gen,
                var_name_gen=var_name_gen,

                additional_temporary_variables={},
                additional_insns=[],
                domains=domains,
                additional_iname_tags={},
                boxed_callables_table=[callables_table],

                inames_added_for_scan=inames_added_for_scan,

                surrounding_within_inames=insn.within_inames,
                surrounding_depends_on=insn.depends_on,
                surrounding_no_sync_with=insn.no_sync_with,
                surrounding_predicates=insn.predicates,

                surrounding_insn_add_within_inames=set(),
                surrounding_insn_add_depends_on=set(),
                surrounding_insn_add_no_sync_with=set(),

                _change_flag=_ChangeFlag(changes_made=False)
                )

        if insn.id not in insn_id_filter \
                or not isinstance(insn, MultiAssignmentBase):
            finished_insns.append(insn)
            continue

        nresults = len(insn.assignees)

        # Run reduction expansion.
        from loopy.symbolic import Reduction
        if isinstance(insn.expression, Reduction) and nresults > 1:
            new_expressions = cb_mapper(insn.expression,
                    red_realize_ctx=red_realize_ctx,
                    nresults=nresults)
        else:
            new_expressions = cb_mapper(insn.expression,
                    red_realize_ctx=red_realize_ctx,
                    nresults=1),

        if red_realize_ctx.were_changes_made:
            # An expansion happened, so insert the generated stuff plus
            # ourselves back into the queue.

            # {{{ apply changes

            kernel_changed = True

            callables_table = red_realize_ctx.boxed_callables_table[0]

            result_assignment_dep_on = (
                    insn.depends_on
                    | frozenset(red_realize_ctx.surrounding_insn_add_depends_on))
            kwargs = insn.get_copy_kwargs(
                    no_sync_with=insn.no_sync_with
                    | frozenset(red_realize_ctx.surrounding_insn_add_no_sync_with),
                    within_inames=(
                        insn.within_inames
                        | red_realize_ctx.surrounding_insn_add_within_inames))

            kwargs.pop("id")
            kwargs.pop("happens_after")
            kwargs.pop("expression")
            kwargs.pop("assignee", None)
            kwargs.pop("assignees", None)
            kwargs.pop("temp_var_type", None)
            kwargs.pop("temp_var_types", None)

            insn_id_replacements = {}

            if isinstance(insn.expression, Reduction) and nresults > 1:
                result_assignment_ids = [
                        insn_id_gen(insn.id) for i in range(nresults)]
                replacement_insns = [
                        Assignment(
                            id=result_assignment_ids[i],
                            depends_on=(
                                result_assignment_dep_on
                                | (frozenset([result_assignment_ids[i-1]])
                                    if i else frozenset())),
                            assignee=assignee,
                            expression=new_expr,
                            **kwargs)
                        for i, (assignee, new_expr) in enumerate(zip(
                            insn.assignees, new_expressions))]

                insn_id_replacements[insn.id] = [
                    rinsn.id for rinsn in replacement_insns]
            else:
                new_expr, = new_expressions
                # since we are replacing the instruction with
                # only one instruction, there's no need to replace id
                replacement_insns = [
                        make_assignment(
                            id=insn.id,
                            depends_on=result_assignment_dep_on,
                            assignees=insn.assignees,
                            expression=new_expr,
                            **kwargs)
                        ]

            additional_insns = red_realize_ctx.additional_insns

            # {{{ make additional insns depend on most recent global barrier

            # FIXME This is weird and hokey and ad-hoc and probably broken.
            # I *think* the idea is to keep a reduction/scan implementation
            # from crossing a global barrier, because that would be costly.

            # check first that the original kernel had global barriers
            # if not, we don't need to check. Since the function
            # kernel_has_global_barriers is cached, we don't do
            # extra work compared to not checking.

            from loopy.kernel.tools import (
                find_most_recent_global_barrier,
                kernel_has_global_barriers,
            )

            if kernel_has_global_barriers(orig_kernel):
                global_barrier = find_most_recent_global_barrier(kernel, insn.id)

                if global_barrier is not None:
                    gb_dep = frozenset([global_barrier])
                    additional_insns = [addl_insn.copy(
                        depends_on=addl_insn.depends_on | gb_dep)
                        for addl_insn in additional_insns]

            # }}}

            # The reduction expander needs an up-to-date kernel
            # object to find dependencies. Keep kernel up-to-date.
            new_temporary_variables = kernel.temporary_variables.copy()
            new_temporary_variables.update(
                    red_realize_ctx.additional_temporary_variables)

            finished_insns = [
                    replace_instruction_ids_in_insn(insn, insn_id_replacements)
                    for insn in finished_insns]
            insn_queue = [
                    replace_instruction_ids_in_insn(insn, insn_id_replacements)
                    for insn in insn_queue]

            finished_insns.extend(additional_insns)
            finished_insns.extend(replacement_insns)

            kernel = kernel.copy(
                    instructions=finished_insns + insn_queue,
                    temporary_variables=new_temporary_variables,
                    domains=domains)
            from loopy.transform.iname import tag_inames
            kernel = tag_inames(kernel, red_realize_ctx.additional_iname_tags)

            del insn_id_replacements

            # }}}

        else:
            # nothing happened, we're done with insn
            assert not red_realize_ctx.surrounding_insn_add_depends_on

            finished_insns.append(insn)

    if not kernel_changed:
        return orig_kernel, callables_table

    kernel = _hackily_ensure_multi_assignment_return_values_are_scoped_private(
                kernel)

    return kernel, callables_table

# }}}


def realize_reduction(t_unit, *args, **kwargs):
    """Rewrites reductions into their imperative form. With *insn_id_filter*
    specified, operate only on the instruction with an instruction id matching
    *insn_id_filter*.

    If *insn_id_filter* is given, only the outermost level of reductions will be
    expanded, inner reductions will be left alone (because they end up in a new
    instruction with a different ID, which doesn't match the filter).

    If *insn_id_filter* is not given, all reductions in all instructions will
    be realized.

    If *automagic_scans_ok*, this function will attempt to rewrite triangular
    reductions as scans automatically.

    If *force_scan* is *True*, this function will attempt to rewrite *all*
    candidate reductions as scans and raise an error if this is not possible
    (this is most useful combined with *insn_id_filter*).

    If *force_outer_iname_for_scan* is not *None*, this function will attempt
    to realize candidate reductions as scans using the specified iname as the
    outer (sweep) iname.

    :arg insn_id_filter: Can be one of:
        - An instance of :class:`str` specifying the ID of the instruction
          whose reductions are to be realized.
        - A collection of :class:`str` specifying the IDs of the instructions
          whose reductions are to be realized.
    """

    assert isinstance(t_unit, TranslationUnit)

    callables_table = dict(t_unit.callables_table)
    kernels_to_scan = [in_knl_callable.subkernel
            for in_knl_callable in t_unit.callables_table.values()
            if isinstance(in_knl_callable, CallableKernel)]

    for knl in kernels_to_scan:
        new_knl, callables_table = realize_reduction_for_single_kernel(
                knl, callables_table, *args, **kwargs)
        in_knl_callable = callables_table[knl.name].copy(
                subkernel=new_knl)
        callables_table[knl.name] = in_knl_callable

    return t_unit.copy(callables_table=constantdict(callables_table))

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2015 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import islpy as isl
import pymbolic.primitives as p
from pymbolic.mapper.differentiator import DifferentiationMapper


var = p.Variable

import loopy as lp
from loopy.diagnostic import LoopyError
from loopy.isl_helpers import make_slab
from loopy.kernel import LoopKernel
from loopy.symbolic import RuleAwareIdentityMapper, SubstitutionRuleMappingContext


# {{{ diff mapper

def func_map(i, func, args, allowed_nonsmoothness):
    if func.name == "exp":
        return var("exp")(*args)
    elif func.name == "log":
        return 1/args[0]

    elif func.name == "sin":
        return var("cos")(*args)
    elif func.name == "cos":
        return -var("sin")(*args)
    elif func.name == "tan":
        return 1+var("tan")(*args)**2

    elif func.name == "sinh":
        return var("cosh")(*args)
    elif func.name == "cosh":
        return var("sinh")(*args)
    elif func.name == "tanh":
        return (1 - var("tanh")(*args))**2

    else:
        raise NotImplementedError("derivative of '%s'" % func.name)


class LoopyDiffMapper(DifferentiationMapper, RuleAwareIdentityMapper):
    def __init__(self, rule_mapping_context, diff_context, diff_inames,
            allowed_nonsmoothness=None):
        RuleAwareIdentityMapper.__init__(self, rule_mapping_context)
        DifferentiationMapper.__init__(
                self,

                # This is actually ignored because we
                # override map_variable below.
                variable=None,

                allowed_nonsmoothness=None)
        self.diff_context = diff_context
        self.diff_inames = diff_inames
        self.diff_iname_exprs = tuple(var(diname) for diname in diff_inames)
        self.function_map = func_map

    def rec_undiff(self, expr, *args):
        dc = self.diff_context

        from loopy.symbolic import get_dependencies
        deps = get_dependencies(expr)
        for dep in deps:
            assert isinstance(dep, str)
            if (dep in dc.kernel.arg_dict
                    or dep in dc.kernel.temporary_variables):
                dc.import_output_var(dep)

        return expr

    def map_variable(self, expr, *args):
        dc = self.diff_context

        if expr.name == dc.by_name:
            assert len(self.diff_inames) == 0
            return 1

        elif (expr.name in dc.kernel.all_inames()
                or expr.name in dc.kernel.all_params()):
            return expr

        else:
            dvar_dby = dc.get_diff_var(expr.name)
            if dvar_dby is None:
                return 0

            if self.diff_inames:
                return var(dvar_dby)[self.diff_iname_exprs]
            else:
                return var(dvar_dby)

    map_tagged_variable = map_variable

    def map_subscript(self, expr, *args):
        dc = self.diff_context

        if expr.aggregate.name == dc.by_name:
            index = expr.index
            if not isinstance(expr.index, tuple):
                index = (expr.index,)

            assert len(self.diff_inames) == len(index)

            conds = [
                p.Comparison(var(ti), "==", ei)
                for ti, ei in zip(self.diff_inames, index)
                ]

            if len(conds) == 1:
                and_conds, = conds
            elif len(conds) > 1:
                and_conds = p.LogicalAnd(tuple(conds))
            else:
                raise AssertionError()

            return p.If(and_conds, 1, 0)

        else:
            dvar_dby = dc.get_diff_var(expr.aggregate.name)
            if dvar_dby is None:
                return 0

            idx = expr.index
            if not isinstance(idx, tuple):
                idx = (idx,)

            return type(expr)(
                    var(dvar_dby),
                    expr.index + self.diff_inames)

    def map_call(self, expr, *args):
        dc = self.diff_context

        if expr.function.name in dc.kernel.substitutions:
            # FIXME: Deal with substitution rules
            # Need to use chain rule here, too.
            raise NotImplementedError("substitution rules in differentiation")
        else:
            return DifferentiationMapper.map_call(self, expr, *args)

# }}}


# {{{ diff context

class DifferentiationContext:
    def __init__(self, kernel, var_name_gen, by_name, diff_iname_prefix,
            additional_shape):
        self.kernel = kernel
        self.by_name = by_name
        self.diff_iname_prefix = diff_iname_prefix
        self.additional_shape = additional_shape

        self.imported_outputs = set()
        self.output_to_diff_output = {}

        self.generate_instruction_id = self.kernel.get_instruction_id_generator()

        self.new_args = []
        self.new_temporary_variables = {}
        self.new_instructions = []
        self.imported_instructions = set()
        self.new_domains = []

        self.rule_mapping_context = SubstitutionRuleMappingContext(
                kernel.substitutions, var_name_gen)

    def get_new_kernel(self):
        knl = self.kernel

        new_args = knl.args + self.new_args
        new_temp_vars = knl.temporary_variables.copy()
        new_temp_vars.update(self.new_temporary_variables)

        knl = knl.copy(
                args=new_args,
                temporary_variables=new_temp_vars,
                instructions=self.new_instructions,
                domains=knl.domains + self.new_domains)

        del new_args
        del new_temp_vars

        knl = self.rule_mapping_context.finish_kernel(knl)

        return knl

    # {{{ kernel gen entrypoints

    def add_diff_inames(self):
        diff_inames = tuple(
            self.rule_mapping_context.make_unique_var_name(
                self.diff_iname_prefix+str(i))
            for i in range(len(self.additional_shape)))

        diff_parameters = set()
        from loopy.symbolic import get_dependencies
        for s in self.additional_shape:
            diff_parameters.update(get_dependencies(s))

        diff_domain = isl.BasicSet(
                "[%s] -> {[%s]}"
                % (", ".join(diff_parameters), ", ".join(diff_inames)))

        for i, diff_iname in enumerate(diff_inames):
            diff_domain = diff_domain & make_slab(
                diff_domain.space, diff_iname, 0, self.additional_shape[i])

        self.new_domains.append(diff_domain)

        return diff_inames

    # }}}

    def import_instruction_and_deps(self, insn_id):
        if insn_id in self.imported_instructions:
            return

        insn = self.kernel.id_to_insn[insn_id]
        self.new_instructions.append(insn)
        self.imported_instructions.add(insn_id)

        id_map = RuleAwareIdentityMapper(self.rule_mapping_context)

        if isinstance(insn, lp.Assignment):
            id_map(insn.expression, self.kernel, insn)
        else:
            raise RuntimeError("do not know how to deal with "
                    "instruction of type %s" % type(insn))

        for dep in insn.depends_on:
            self.import_instruction_and_deps(dep)

    def import_output_var(self, var_name):
        writers = self.kernel.writer_map().get(var_name, [])

        if len(writers) > 1:
            raise LoopyError("%s is written in more than one place"
                    % var_name)

        if not writers:
            return

        insn_id, = writers
        self.import_instruction_and_deps(insn_id)

    def get_diff_var(self, var_name):
        """
        :return: a string containing the name of a new variable
            holding the derivative of *var_name* by the desired
            *diff_context.by_name*, or *None* if no dependency exists.
        """
        new_var_name = self.rule_mapping_context.make_unique_var_name(
                var_name + "_d" + self.by_name)

        writers = self.kernel.writer_map().get(var_name, [])

        if not writers:
            # FIXME: There should be hooks to supply earlier dvar_dby
            # This would be the spot to think about them.
            return None

        if len(writers) > 1:
            raise LoopyError("%s is written in more than one place"
                    % var_name)

        orig_writer_id, = writers
        orig_writer_insn = self.kernel.id_to_insn[orig_writer_id]

        diff_inames = self.add_diff_inames()
        diff_iname_exprs = tuple(var(diname) for diname in diff_inames)

        # {{{ write code

        diff_mapper = LoopyDiffMapper(self.rule_mapping_context, self,
                diff_inames)

        diff_expr = diff_mapper(orig_writer_insn.expression,
                self.kernel, orig_writer_insn)

        if not diff_expr:
            return None

        assert isinstance(orig_writer_insn, lp.Assignment)
        if isinstance(orig_writer_insn.assignee, p.Subscript):
            lhs_ind = orig_writer_insn.assignee.index_tuple
        elif isinstance(orig_writer_insn.assignee, p.Variable):
            lhs_ind = ()
        else:
            raise LoopyError(
                    "Unrecognized LHS type in differentiation: %s"
                    % type(orig_writer_insn.assignee).__name__)

        new_insn_id = self.generate_instruction_id()
        insn = lp.Assignment(
                id=new_insn_id,
                assignee=var(new_var_name)[
                    lhs_ind + diff_iname_exprs],
                expression=diff_expr,
                within_inames=(
                    orig_writer_insn.within_inames | frozenset(diff_inames)))

        self.new_instructions.append(insn)

        # }}}

        # {{{ manage variable declaration

        if var_name in self.kernel.arg_dict:
            arg = self.kernel.arg_dict[var_name]
            orig_shape = arg.shape

        elif var_name in self.kernel.temporary_variables:
            tv = self.kernel.temporary_variables[var_name]
            orig_shape = tv.shape

        else:
            raise ValueError("%s: variable not found" % var_name)

        shape = orig_shape + self.additional_shape
        dim_tags = ("c",) * len(shape)

        if var_name in self.kernel.arg_dict:
            self.new_args.append(
                lp.GlobalArg(
                    new_var_name,
                    arg.dtype,
                    shape=shape,
                    dim_tags=dim_tags,
                    is_input=arg.is_input,
                    is_output=arg.is_output
                ))

        elif var_name in self.kernel.temporary_variables:
            self.new_temporary_variables[new_var_name] = lp.TemporaryVariable(
                    new_var_name,
                    tv.dtype,
                    shape=shape,
                    dim_tags=dim_tags)

        # }}}

        return new_var_name

# }}}


# {{{ entrypoint

def diff_kernel(kernel, diff_outputs, by, diff_iname_prefix="diff_i",
        batch_axes_in_by=frozenset(), copy_outputs=frozenset()):
    """

    :arg batch_axes_in_by: a :class:`set` of axis indices in the variable named *by*
        that are not part of the differentiation.
    :return: a string containing the name of a new variable
        holding the derivative of *var_name* by the desired
        *diff_context.by_name*, or *None* if no dependency exists.
    """

    assert isinstance(kernel, LoopKernel)

    from loopy.kernel.creation import apply_single_writer_dependency_heuristic
    kernel = apply_single_writer_dependency_heuristic(kernel, warn_if_used=True)

    if isinstance(diff_outputs, str):
        diff_outputs = [
                d_out.strip() for d_out in diff_outputs.split(",")
                if d_out.strip()]

    by_arg = kernel.arg_dict[by]
    additional_shape = by_arg.shape

    var_name_gen = kernel.get_var_name_generator()

    # {{{ differentiate instructions

    diff_context = DifferentiationContext(
            kernel, var_name_gen, by, diff_iname_prefix=diff_iname_prefix,
            additional_shape=additional_shape)

    result = {}
    for d_out in diff_outputs:
        result = diff_context.get_diff_var(d_out)

    for cout in copy_outputs:
        diff_context.import_output_var(cout)

    # }}}

    return diff_context.get_new_kernel(), result

# }}}


# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from pytools import MovedFunctionDeprecationWrapper

from loopy.diagnostic import LoopyError
from loopy.kernel import LoopKernel
from loopy.kernel.function_interface import CallableKernel
from loopy.symbolic import RuleAwareIdentityMapper, SubstitutionRuleMappingContext
from loopy.translation_unit import TranslationUnit, for_each_kernel


class SubscriptRewriter(RuleAwareIdentityMapper):
    def __init__(self, rule_mapping_context, arg_names, handler):
        super().__init__(rule_mapping_context)
        self.arg_names = arg_names
        self.handler = handler

    def map_subscript(self, expr, expn_state):
        if expr.aggregate.name in self.arg_names:
            return self.handler(expr)
        else:
            return super().map_subscript(expr, expn_state)

    def map_kernel(self, kernel, within=lambda *args: True):
        new_insns = [
            # While subst rules are not allowed in assignees, the mapper
            # may perform tasks entirely unrelated to subst rules, so
            # we must map assignees, too.
            insn if not kernel.substitutions and not within(kernel, insn, ())
            and not any(name in self.arg_names for name in insn.dependency_names())
            else self.map_instruction(kernel,
                insn.with_transformed_expressions(
                    lambda expr: self(expr, kernel, insn)))  # noqa: B023
            for insn in kernel.instructions]

        return kernel.copy(instructions=new_insns)


# {{{ split_array_dim (deprecated since June 2016)

@for_each_kernel
def split_array_dim(kernel, arrays_and_axes, count,
        auto_split_inames=True,
        split_kwargs=None):
    """
    :arg arrays_and_axes: a list of tuples *(array, axis_nr)* indicating
        that the index in *axis_nr* should be split. The tuples may
        also be *(array, axis_nr, "F")*, indicating that the index will
        be split as it would be according to Fortran order.

        *array* may name a temporary variable or an argument.

        If *arrays_and_axes* is a :class:`tuple`, it is automatically
        wrapped in a list, to make single splits easier.

    :arg count: The group size to use in the split.
    :arg auto_split_inames: Whether to automatically split inames
        encountered in the specified indices.
    :arg split_kwargs: arguments to pass to :func:`loopy.split_inames`

    Note that splits on the corresponding inames are carried out implicitly.
    The inames may *not* be split beforehand. (There's no *really* good reason
    for this--this routine is just not smart enough to deal with this.)
    """

    if count == 1:
        return kernel

    if split_kwargs is None:
        split_kwargs = {}

    # {{{ process input into array_to_rest

    # where "rest" is the non-argument-name part of the input tuples
    # in args_and_axes
    def normalize_rest(rest):
        if len(rest) == 1:
            return (rest[0], "C")
        elif len(rest) == 2:
            return rest
        else:
            raise RuntimeError("split instruction '%s' not understood" % rest)

    if isinstance(arrays_and_axes, tuple):
        arrays_and_axes = [arrays_and_axes]

    array_to_rest = {
            tup[0]: normalize_rest(tup[1:]) for tup in arrays_and_axes}

    if len(arrays_and_axes) != len(array_to_rest):
        raise RuntimeError("cannot split multiple axes of the same variable")

    del arrays_and_axes

    # }}}

    # {{{ adjust arrays

    from loopy.kernel.tools import ArrayChanger

    for array_name, (axis, order) in array_to_rest.items():
        achng = ArrayChanger(kernel, array_name)
        ary = achng.get()

        from pytools import div_ceil

        # {{{ adjust shape

        new_shape = ary.shape
        if new_shape is not None:
            new_shape = list(new_shape)
            axis_len = new_shape[axis]
            new_shape[axis] = count
            outer_len = div_ceil(axis_len, count)

            if order == "F":
                new_shape.insert(axis+1, outer_len)
            elif order == "C":
                new_shape.insert(axis, outer_len)
            else:
                raise RuntimeError("order '%s' not understood" % order)
            new_shape = tuple(new_shape)

        # }}}

        # {{{ adjust dim tags

        if ary.dim_tags is None:
            raise RuntimeError("dim_tags of '%s' are not known" % array_name)
        new_dim_tags = list(ary.dim_tags)

        old_dim_tag = ary.dim_tags[axis]

        from loopy.kernel.array import FixedStrideArrayDimTag
        if not isinstance(old_dim_tag, FixedStrideArrayDimTag):
            raise RuntimeError("axis %d of '%s' is not tagged fixed-stride"
                    % (axis, array_name))

        old_stride = old_dim_tag.stride
        outer_stride = count*old_stride

        if order == "F":
            new_dim_tags.insert(axis+1, FixedStrideArrayDimTag(outer_stride))
        elif order == "C":
            new_dim_tags.insert(axis, FixedStrideArrayDimTag(outer_stride))
        else:
            raise RuntimeError("order '%s' not understood" % order)

        new_dim_tags = tuple(new_dim_tags)

        # }}}

        # {{{ adjust dim_names

        new_dim_names = ary.dim_names
        if new_dim_names is not None:
            new_dim_names = list(new_dim_names)
            existing_name = new_dim_names[axis]
            new_dim_names[axis] = existing_name + "_inner"
            outer_name = existing_name + "_outer"

            if order == "F":
                new_dim_names.insert(axis+1, outer_name)
            elif order == "C":
                new_dim_names.insert(axis, outer_name)
            else:
                raise RuntimeError("order '%s' not understood" % order)
            new_dim_names = tuple(new_dim_names)

        # }}}

        kernel = achng.with_changed_array(ary.copy(
            shape=new_shape, dim_tags=new_dim_tags, dim_names=new_dim_names))

    # }}}

    split_vars = {}

    var_name_gen = kernel.get_var_name_generator()

    def split_access_axis(expr):
        axis_nr, order = array_to_rest[expr.aggregate.name]

        idx = expr.index
        if not isinstance(idx, tuple):
            idx = (idx,)
        idx = list(idx)

        axis_idx = idx[axis_nr]

        if auto_split_inames:
            from pymbolic.primitives import Variable
            if not isinstance(axis_idx, Variable):
                raise RuntimeError("found access '%s' in which axis %d is not a "
                        "single variable--cannot split "
                        "(Have you tried to do the split yourself, manually, "
                        "beforehand? If so, you shouldn't.)"
                        % (expr, axis_nr))

            split_iname = idx[axis_nr].name
            assert split_iname in kernel.all_inames()

            try:
                outer_iname, inner_iname = split_vars[split_iname]
            except KeyError:
                outer_iname = var_name_gen(split_iname+"_outer")
                inner_iname = var_name_gen(split_iname+"_inner")
                split_vars[split_iname] = outer_iname, inner_iname

            inner_index = Variable(inner_iname)
            outer_index = Variable(outer_iname)

        else:
            from loopy.symbolic import simplify_using_aff
            inner_index = simplify_using_aff(kernel, axis_idx % count)
            outer_index = simplify_using_aff(kernel, axis_idx // count)

        idx[axis_nr] = inner_index

        if order == "F":
            idx.insert(axis+1, outer_index)
        elif order == "C":
            idx.insert(axis, outer_index)
        else:
            raise RuntimeError("order '%s' not understood" % order)

        return expr.aggregate.index(tuple(idx))

    rule_mapping_context = SubstitutionRuleMappingContext(
            kernel.substitutions, var_name_gen)
    aash = SubscriptRewriter(rule_mapping_context,
            set(array_to_rest.keys()), split_access_axis)
    kernel = rule_mapping_context.finish_kernel(aash.map_kernel(kernel))

    if auto_split_inames:
        from loopy import split_iname
        for iname, (outer_iname, inner_iname) in split_vars.items():
            kernel = split_iname(kernel, iname, count,
                    outer_iname=outer_iname, inner_iname=inner_iname,
                    **split_kwargs)

    return kernel


split_arg_axis = (MovedFunctionDeprecationWrapper(split_array_dim))

# }}}


# {{{ split_array_axis

def _split_array_axis_inner(kernel, array_name, axis_nr, count, order="C"):
    if count == 1:
        return kernel

    # {{{ adjust arrays

    from loopy.kernel.tools import ArrayChanger

    achng = ArrayChanger(kernel, array_name)
    ary = achng.get()

    from pytools import div_ceil

    # {{{ adjust shape

    new_shape = ary.shape
    if new_shape is not None:
        new_shape = list(new_shape)
        axis_len = new_shape[axis_nr]
        new_shape[axis_nr] = count
        outer_len = div_ceil(axis_len, count)

        if order == "F":
            new_shape.insert(axis_nr+1, outer_len)
        elif order == "C":
            new_shape.insert(axis_nr, outer_len)
        else:
            raise RuntimeError("order '%s' not understood" % order)
        new_shape = tuple(new_shape)

    # }}}

    # {{{ adjust dim tags

    if ary.dim_tags is None:
        raise RuntimeError("dim_tags of '%s' are not known" % array_name)
    new_dim_tags = list(ary.dim_tags)

    old_dim_tag = ary.dim_tags[axis_nr]

    from loopy.kernel.array import FixedStrideArrayDimTag
    if not isinstance(old_dim_tag, FixedStrideArrayDimTag):
        raise RuntimeError("axis %d of '%s' is not tagged fixed-stride"
                % (axis_nr, array_name))

    old_stride = old_dim_tag.stride
    outer_stride = count*old_stride

    if order == "F":
        new_dim_tags.insert(axis_nr+1, FixedStrideArrayDimTag(outer_stride))
    elif order == "C":
        new_dim_tags.insert(axis_nr, FixedStrideArrayDimTag(outer_stride))
    else:
        raise RuntimeError("order '%s' not understood" % order)

    new_dim_tags = tuple(new_dim_tags)

    # }}}

    # {{{ adjust dim_names

    new_dim_names = ary.dim_names
    if new_dim_names is not None:
        new_dim_names = list(new_dim_names)
        existing_name = new_dim_names[axis_nr]
        new_dim_names[axis_nr] = existing_name + "_inner"
        outer_name = existing_name + "_outer"

        if order == "F":
            new_dim_names.insert(axis_nr+1, outer_name)
        elif order == "C":
            new_dim_names.insert(axis_nr, outer_name)
        else:
            raise RuntimeError("order '%s' not understood" % order)
        new_dim_names = tuple(new_dim_names)

    # }}}

    kernel = achng.with_changed_array(ary.copy(
        shape=new_shape, dim_tags=new_dim_tags, dim_names=new_dim_names))

    # }}}

    var_name_gen = kernel.get_var_name_generator()

    def split_access_axis(expr):
        idx = expr.index
        if not isinstance(idx, tuple):
            idx = (idx,)
        idx = list(idx)

        axis_idx = idx[axis_nr]

        from loopy.symbolic import simplify_using_aff
        inner_index = simplify_using_aff(kernel, axis_idx % count)
        outer_index = simplify_using_aff(kernel, axis_idx // count)

        idx[axis_nr] = inner_index

        if order == "F":
            idx.insert(axis_nr+1, outer_index)
        elif order == "C":
            idx.insert(axis_nr, outer_index)
        else:
            raise RuntimeError("order '%s' not understood" % order)

        return expr.aggregate[tuple(idx)]

    rule_mapping_context = SubstitutionRuleMappingContext(
            kernel.substitutions, var_name_gen)
    aash = SubscriptRewriter(rule_mapping_context,
            {array_name}, split_access_axis)
    kernel = rule_mapping_context.finish_kernel(aash.map_kernel(kernel))

    return kernel


@for_each_kernel
def split_array_axis(kernel, array_names, axis_nr, count,
        order="C"):
    """
    :arg array: a list of names of temporary variables or arguments. May
        also be a comma-separated string of these.

    :arg axis_nr: the (zero-based) index of the axis that should be split.

    :arg count: The group size to use in the split.

    :arg order: The way the split array axis should be linearized.
        May be "C" or "F" to indicate C/Fortran (row/column)-major order.

    .. versionchanged:: 2016.2

        There was a more complicated, dumber function called
        ``loopy.split_array_dim`` that had the role of this function in
        versions prior to 2016.2.
    """
    assert isinstance(kernel, LoopKernel)

    if isinstance(array_names, str):
        array_names = [i.strip() for i in array_names.split(",") if i.strip()]

    for array_name in array_names:
        kernel = _split_array_axis_inner(kernel, array_name, axis_nr, count, order)

    return kernel

# }}}


# {{{ find_padding_multiple

def find_padding_multiple(kernel, variable, axis, align_bytes, allowed_waste=0.1):
    if isinstance(kernel, TranslationUnit):
        kernel_names = [i for i, clbl in kernel.callables_table.items()
                if isinstance(clbl, CallableKernel)]
        if len(kernel_names) > 1:
            raise LoopyError()
        return find_padding_multiple(kernel[kernel_names[0]], variable, axis,
                align_bytes, allowed_waste)
    assert isinstance(kernel, LoopKernel)

    arg = kernel.arg_dict[variable]

    if arg.dim_tags is None:
        raise RuntimeError("cannot find padding multiple--dim_tags of '%s' "
                "are not known" % variable)

    dim_tag = arg.dim_tags[axis]

    from loopy.kernel.array import FixedStrideArrayDimTag
    if not isinstance(dim_tag, FixedStrideArrayDimTag):
        raise RuntimeError("cannot find padding multiple--"
                "axis %d of '%s' is not tagged fixed-stride"
                % (axis, variable))

    stride = dim_tag.stride

    if not isinstance(stride, int):
        raise RuntimeError("cannot find padding multiple--stride is not a "
                "known integer")

    from pytools import div_ceil

    multiple = 1
    while True:
        true_size = multiple * stride
        padded_size = div_ceil(true_size, align_bytes) * align_bytes

        if (padded_size - true_size) / true_size <= allowed_waste:
            return multiple

        multiple += 1

# }}}


# {{{ add_padding

@for_each_kernel
def add_padding(kernel, variable, axis, align_bytes):
    arg_to_idx = {arg.name: i for i, arg in enumerate(kernel.args)}
    arg_idx = arg_to_idx[variable]

    new_args = kernel.args[:]
    arg = new_args[arg_idx]

    if arg.dim_tags is None:
        raise RuntimeError("cannot add padding--dim_tags of '%s' "
                "are not known" % variable)

    new_dim_tags = list(arg.dim_tags)
    dim_tag = new_dim_tags[axis]

    from loopy.kernel.array import FixedStrideArrayDimTag
    if not isinstance(dim_tag, FixedStrideArrayDimTag):
        raise RuntimeError("cannot find padding multiple--"
                "axis %d of '%s' is not tagged fixed-stride"
                % (axis, variable))

    stride = dim_tag.stride
    if not isinstance(stride, int):
        raise RuntimeError("cannot find split granularity--stride is not a "
                "known integer")

    from pytools import div_ceil
    new_dim_tags[axis] = FixedStrideArrayDimTag(
            div_ceil(stride, align_bytes) * align_bytes)

    new_args[arg_idx] = arg.copy(dim_tags=tuple(new_dim_tags))

    return kernel.copy(args=new_args)

# }}}


# vim: foldmethod=marker



import numpy as np

import loopy as lp
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


knl = lp.make_kernel(
        "{ [i,k]: 0<=i<n and 0<=k<3 }",
        """
        for i, k
            ... gbarrier
            c[k,i] = a[k, i + 1]
            ... gbarrier
            out[k,i] = c[k,i]
        end
        """, seq_dependencies=True)

# transform
knl = lp.split_iname(knl, "i", 128, outer_tag="g.0", inner_tag="l.0")
knl = lp.add_and_infer_dtypes(knl,
        {"a": np.float32, "c": np.float32, "out": np.float32, "n": np.int32})

# schedule
from loopy.preprocess import preprocess_kernel


knl = preprocess_kernel(knl)

from loopy.schedule import get_one_linearized_kernel


knl = knl.with_kernel(get_one_linearized_kernel(knl["loopy_kernel"],
                                                knl.callables_table))


# map schedule onto host or device
print(knl)

cgr = lp.generate_code_v2(knl)

print(cgr.device_code())
print(cgr.host_code())



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from dataclasses import dataclass
from typing import TYPE_CHECKING
from warnings import warn

import numpy as np

import loopy as lp
from loopy.diagnostic import AutomaticTestFailure, LoopyError
from loopy.kernel.array import get_strides


if TYPE_CHECKING:
    import pyopencl.array as cla


AUTO_TEST_SKIP_RUN = False

import logging


logger = logging.getLogger(__name__)


def is_dtype_supported(dtype):
    # Work around https://github.com/numpy/numpy/issues/4317
    return dtype.kind in "biufc"


def evaluate_shape(shape, context):
    from pymbolic import evaluate

    result = []
    for saxis in shape:
        if saxis is None:
            result.append(saxis)
        else:
            result.append(evaluate(saxis, context))

    return tuple(result)


# {{{ create random argument arrays for testing

def fill_rand(ary):
    from pyopencl.clrandom import fill_rand
    if ary.dtype.kind == "c":
        real_dtype = ary.dtype.type(0).real.dtype
        real_ary = ary.view(real_dtype)

        fill_rand(real_ary)
    else:
        fill_rand(ary)


@dataclass
class TestArgInfo:
    name: str
    ref_array: cla.Array
    ref_storage_array: cla.Array

    ref_pre_run_array: cla.Array
    ref_pre_run_storage_array: cla.Array

    ref_shape: tuple[int, ...]
    ref_strides: tuple[int, ...]
    ref_alloc_size: int
    ref_numpy_strides: tuple[int, ...]
    needs_checking: bool

    # The attributes below are being modified in make_args, hence this dataclass
    # cannot be frozen.
    test_storage_array: cla.Array | None = None
    test_array: cla.Array | None = None
    test_shape: tuple[int, ...] | None = None
    test_strides: tuple[int, ...] | None = None
    test_numpy_strides: tuple[int, ...] | None = None
    test_alloc_size: tuple[int, ...] | None = None


# {{{ "reference" arguments

def make_ref_args(kernel, queue, parameters):
    import pyopencl as cl
    import pyopencl.array as cl_array
    from pymbolic import evaluate

    from loopy.kernel.data import (
        ArrayArg,
        ConstantArg,
        ImageArg,
        TemporaryVariable,
        ValueArg,
    )

    ref_args = {}
    ref_arg_data = []

    for arg in kernel.args:
        if isinstance(arg, ValueArg):
            arg_value = parameters[arg.name]

            try:
                argv_dtype = arg_value.dtype
            except AttributeError:
                argv_dtype = None

            if argv_dtype != arg.dtype:
                arg_value = arg.dtype.numpy_dtype.type(arg_value)

            ref_args[arg.name] = arg_value

            ref_arg_data.append(None)

        elif isinstance(arg, (ArrayArg, ImageArg, ConstantArg)):
            if arg.shape is None or any(saxis is None for saxis in arg.shape):
                raise LoopyError("array '%s' needs known shape to use automatic "
                        "testing" % arg.name)

            shape = evaluate_shape(arg.shape, parameters)
            dtype = arg.dtype

            is_output = arg.is_output

            if isinstance(arg, ImageArg):
                storage_array = ary = cl_array.empty(
                        queue, shape, dtype, order="C")
                numpy_strides = None
                alloc_size = None
                strides = None
            else:
                strides = evaluate(get_strides(arg), parameters)

                alloc_size = sum(astrd*(alen-1) if astrd != 0 else alen-1
                        for alen, astrd in zip(shape, strides)) + 1

                if dtype is None:
                    raise LoopyError("dtype for argument '%s' is not yet "
                            "known. Perhaps you want to use "
                            "loopy.add_dtypes "
                            "or loopy.infer_argument_dtypes?"
                            % arg.name)

                itemsize = dtype.itemsize
                numpy_strides = [itemsize*s for s in strides]

                storage_array = cl_array.empty(queue, alloc_size, dtype)

            if is_output and isinstance(arg, ImageArg):
                raise LoopyError("write-mode images not supported in "
                        "automatic testing")

            fill_rand(storage_array)

            if isinstance(arg, ImageArg):
                # must be contiguous
                pre_run_ary = pre_run_storage_array = storage_array.copy()

                ref_args[arg.name] = cl.image_from_array(
                        queue.context, ary.get())
            else:
                pre_run_storage_array = storage_array.copy()

                ary = cl_array.as_strided(storage_array, shape, numpy_strides)
                pre_run_ary = cl_array.as_strided(
                        pre_run_storage_array, shape, numpy_strides)
                ref_args[arg.name] = ary

            ref_arg_data.append(
                    TestArgInfo(
                        name=arg.name,
                        ref_array=ary,
                        ref_storage_array=storage_array,

                        ref_pre_run_array=pre_run_ary,
                        ref_pre_run_storage_array=pre_run_storage_array,

                        ref_shape=shape,
                        ref_strides=strides,
                        ref_alloc_size=alloc_size,
                        ref_numpy_strides=numpy_strides,
                        needs_checking=is_output))

        elif arg.arg_class is TemporaryVariable:
            # global temporary, handled by invocation logic
            pass

        else:
            raise LoopyError("arg type %s not understood" % type(arg))

    return ref_args, ref_arg_data

# }}}


# {{{ "full-scale" arguments

def make_args(kernel, queue, ref_arg_data, parameters):
    import pyopencl as cl
    import pyopencl.array as cl_array
    from pymbolic import evaluate

    from loopy.kernel.data import ArrayArg, ConstantArg, ImageArg, ValueArg

    args = {}
    for arg, arg_desc in zip(kernel.args, ref_arg_data):
        if isinstance(arg, ValueArg):
            arg_value = parameters[arg.name]

            try:
                argv_dtype = arg_value.dtype
            except AttributeError:
                argv_dtype = None

            if argv_dtype != arg.dtype:
                arg_value = arg.dtype.numpy_dtype.type(arg_value)

            args[arg.name] = arg_value

        elif isinstance(arg, ImageArg):
            if arg.name in kernel.get_written_variables():
                raise NotImplementedError("write-mode images not supported in "
                        "automatic testing")

            shape = evaluate_shape(arg.shape, parameters)
            assert shape == arg_desc.ref_shape

            # must be contiguous
            args[arg.name] = cl.image_from_array(
                    queue.context, arg_desc.ref_pre_run_array.get())

        elif isinstance(arg, (ArrayArg, ConstantArg)):
            shape = evaluate(arg.shape, parameters)
            strides = evaluate(get_strides(arg), parameters)

            dtype = arg.dtype
            itemsize = dtype.itemsize
            numpy_strides = [itemsize*s for s in strides]

            alloc_size = sum(astrd*(alen-1) if astrd != 0 else alen-1
                    for alen, astrd in zip(shape, strides)) + 1

            # use contiguous array to transfer to host
            host_ref_contig_array = arg_desc.ref_pre_run_storage_array.get()

            # use device shape/strides
            from pyopencl.compyte.array import as_strided
            host_ref_array = as_strided(host_ref_contig_array,
                    arg_desc.ref_shape, arg_desc.ref_numpy_strides)

            # flatten the thing
            host_ref_flat_array = host_ref_array.flatten()

            # create host array with test shape (but not strides)
            host_contig_array = np.empty(shape, dtype=dtype)

            common_len = min(
                    len(host_ref_flat_array),
                    len(host_contig_array.ravel()))
            host_contig_array.ravel()[:common_len] = \
                    host_ref_flat_array[:common_len]

            # create host array with test shape and storage layout
            host_storage_array = np.empty(alloc_size, dtype)
            host_array = as_strided(
                    host_storage_array, shape, numpy_strides)
            host_array[...] = host_contig_array

            host_contig_array = arg_desc.ref_storage_array.get()
            storage_array = cl_array.to_device(queue, host_storage_array)
            ary = cl_array.as_strided(storage_array, shape, numpy_strides)

            args[arg.name] = ary

            arg_desc.test_storage_array = storage_array
            arg_desc.test_array = ary
            arg_desc.test_shape = shape
            arg_desc.test_strides = strides
            arg_desc.test_numpy_strides = numpy_strides
            arg_desc.test_alloc_size = alloc_size

        else:
            raise LoopyError("arg type not understood")

    return args

# }}}

# }}}


# {{{ default array comparison

def _default_check_result(result, ref_result):
    if not is_dtype_supported(result.dtype) and not (result == ref_result).all():
        return (False, "results do not match exactly")

    if not np.allclose(ref_result, result, rtol=1e-3, atol=1e-3):
        l2_err = (
                np.sum(np.abs(ref_result-result)**2)
                / np.sum(np.abs(ref_result)**2))
        linf_err = (
                np.max(np.abs(ref_result-result))
                / np.max(np.abs(ref_result-result)))
        # pylint: disable=bad-string-format-type
        return (False,
                # pylint: disable=bad-string-format-type
                "results do not match -- (rel) l_2 err: %g, l_inf err: %g"
                % (l2_err, linf_err))
    else:
        return True, None

# }}}


# {{{ find device for reference test

def _enumerate_cl_devices_for_ref_test(blacklist_ref_vendors, need_image_support):
    import pyopencl as cl

    noncpu_devs = []
    cpu_devs = []

    if isinstance(blacklist_ref_vendors, str):
        blacklist_ref_vendors = blacklist_ref_vendors.split(",")

    for pf in cl.get_platforms():
        for dev in pf.get_devices():
            if any(bl in dev.platform.vendor
                    for bl in blacklist_ref_vendors):
                continue

            if need_image_support:
                if not dev.image_support:
                    continue
                if pf.vendor == "The pocl project":
                    # Hahaha, no.
                    continue

            if dev.type & cl.device_type.CPU:
                cpu_devs.append(dev)
            else:
                noncpu_devs.append(dev)

    if not (cpu_devs or noncpu_devs):
        raise LoopyError("no CL device found for test")

    if not cpu_devs:
        warn("No CPU device found for running reference kernel. The reference "
                "computation will either fail because of a timeout "
                "or take a *very* long time.", stacklevel=1)

    for dev in cpu_devs:
        yield dev

    for dev in noncpu_devs:
        yield dev

# }}}


# {{{ main automatic testing entrypoint

def auto_test_vs_ref(
        ref_prog, ctx, test_prog=None, op_count=(), op_label=(), parameters=None,
        print_ref_code=False, print_code=True, warmup_rounds=2,
        dump_binary=False,
        fills_entire_output=None, do_check=True, check_result=None,
        max_test_kernel_count=1,
        quiet=False, blacklist_ref_vendors=(), ref_entrypoint=None,
        test_entrypoint=None):
    """Compare results of `ref_knl` to the kernels generated by
    scheduling *test_knl*.

    :arg check_result: a callable with :class:`numpy.ndarray` arguments
        *(result, reference_result)* returning a a tuple (class:`bool`,
        message) indicating correctness/acceptability of the result
    :arg max_test_kernel_count: Stop testing after this many *test_knl*
    """
    if parameters is None:
        parameters = {}

    import pyopencl as cl

    if test_prog is None:
        test_prog = ref_prog
        do_check = False

    if ref_entrypoint is None:
        if len(ref_prog.entrypoints) != 1:
            raise LoopyError("Unable to guess entrypoint for ref_prog.")
        ref_entrypoint = next(iter(ref_prog.entrypoints))

    if test_entrypoint is None:
        if len(test_prog.entrypoints) != 1:
            raise LoopyError("Unable to guess entrypoint for ref_prog.")
        test_entrypoint = next(iter(test_prog.entrypoints))

    ref_prog = lp.preprocess_kernel(ref_prog)
    test_prog = lp.preprocess_kernel(test_prog)

    if len(ref_prog[ref_entrypoint].args) != len(test_prog[test_entrypoint].args):
        raise LoopyError("ref_prog and test_prog do not have the same number "
                "of arguments")

    for i, (ref_arg, test_arg) in enumerate(zip(ref_prog[ref_entrypoint].args,
            test_prog[test_entrypoint].args)):
        if ref_arg.name != test_arg.name:
            raise LoopyError("ref_prog and test_prog argument lists disagree at "
                    "index %d (1-based)" % (i+1))

        if ref_arg.dtype != test_arg.dtype:
            raise LoopyError("ref_prog and test_prog argument lists disagree at "
                    "index %d (1-based)" % (i+1))

    from loopy.target.execution import get_highlighted_code

    if isinstance(op_count, (int, float)):
        warn("op_count should be a list", stacklevel=2)
        op_count = [op_count]
    if isinstance(op_label, str):
        warn("op_label should be a list", stacklevel=2)
        op_label = [op_label]

    from time import time

    if check_result is None:
        check_result = _default_check_result

    if fills_entire_output is not None:
        warn("fills_entire_output is deprecated", DeprecationWarning, stacklevel=2)

    # {{{ compile and run reference code

    from loopy.type_inference import infer_unknown_types
    ref_prog = infer_unknown_types(ref_prog, expect_completion=True)

    found_ref_device = False

    ref_errors = []

    from loopy.kernel.data import ImageArg
    need_ref_image_support = any(isinstance(arg, ImageArg)
                                 for arg in ref_prog[ref_entrypoint].args)

    for dev in _enumerate_cl_devices_for_ref_test(
            blacklist_ref_vendors, need_ref_image_support):

        ref_ctx = cl.Context([dev])
        ref_queue = cl.CommandQueue(ref_ctx,
                properties=cl.command_queue_properties.PROFILING_ENABLE)
        ref_codegen_result = lp.generate_code_v2(ref_prog)

        logger.info("{} (ref): trying {} for the reference calculation".format(
            ref_entrypoint, dev))

        if not quiet and print_ref_code:
            print(75*"-")
            print("Reference Code:")
            print(75*"-")
            print(get_highlighted_code(
                ref_codegen_result.device_code()))
            print(75*"-")

        try:
            ref_args, ref_arg_data = \
                    make_ref_args(ref_prog[ref_entrypoint], ref_queue, parameters)
            ref_args["out_host"] = False
        except cl.RuntimeError as e:
            if e.code == cl.status_code.IMAGE_FORMAT_NOT_SUPPORTED:
                import traceback
                ref_errors.append("\n".join([
                    75*"-",
                    "On %s:" % dev,
                    75*"-",
                    traceback.format_exc(),
                    75*"-"]))

                continue
            else:
                raise

        found_ref_device = True

        if not do_check:
            break

        ref_queue.finish()

        logger.info("{} (ref): using {} for the reference calculation".format(
            ref_entrypoint, dev))
        logger.info("%s (ref): run" % ref_entrypoint)

        ref_start = time()

        if not AUTO_TEST_SKIP_RUN:
            ref_evt, _ = ref_prog(ref_queue, **ref_args)
        else:
            ref_evt = cl.enqueue_marker(ref_queue)

        ref_queue.finish()
        ref_stop = time()
        ref_elapsed_wall = ref_stop-ref_start

        logger.info("%s (ref): run done" % ref_entrypoint)

        ref_evt.wait()
        ref_elapsed_event = 1e-9*(ref_evt.profile.END-ref_evt.profile.START)

        break

    if not found_ref_device:
        raise LoopyError("could not find a suitable device for the "
                "reference computation.\n"
                "These errors were encountered:\n"+"\n".join(ref_errors))

    # }}}

    # {{{ compile and run parallel code

    need_check = do_check

    queue = cl.CommandQueue(ctx,
            properties=cl.command_queue_properties.PROFILING_ENABLE)

    from loopy.kernel import KernelState
    from loopy.target.pyopencl import PyOpenCLTarget
    if test_prog[test_entrypoint].state not in [
            KernelState.PREPROCESSED,
            KernelState.LINEARIZED]:
        if isinstance(test_prog.target, PyOpenCLTarget):
            test_prog = test_prog.copy(target=PyOpenCLTarget(ctx.devices[0]))

        test_prog = lp.preprocess_kernel(test_prog)

    from loopy.type_inference import infer_unknown_types

    test_prog = infer_unknown_types(test_prog, expect_completion=True)
    test_prog_codegen_result = lp.generate_code_v2(test_prog)

    args = make_args(test_prog[test_entrypoint],
            queue, ref_arg_data, parameters)
    args["out_host"] = False

    if not quiet:
        print(75*"-")
        print("Kernel:")
        print(75*"-")
        if print_code:
            print(get_highlighted_code(
                test_prog_codegen_result.device_code()))
            print(75*"-")
        if dump_binary:
            print(type(test_prog_codegen_result.cl_program))
            print(test_prog_codegen_result.cl_program.binaries[0])
            print(75*"-")

    logger.info("%s: run warmup" % (test_entrypoint))

    for _i in range(warmup_rounds):
        if not AUTO_TEST_SKIP_RUN:
            test_prog(queue, **args)

        if need_check and not AUTO_TEST_SKIP_RUN:
            for arg_desc in ref_arg_data:
                if arg_desc is None:
                    continue
                if not arg_desc.needs_checking:
                    continue

                from pyopencl.compyte.array import as_strided
                ref_ary = as_strided(
                        arg_desc.ref_storage_array.get(),
                        shape=arg_desc.ref_shape,
                        strides=arg_desc.ref_numpy_strides).flatten()
                test_ary = as_strided(
                        arg_desc.test_storage_array.get(),
                        shape=arg_desc.test_shape,
                        strides=arg_desc.test_numpy_strides).flatten()
                common_len = min(len(ref_ary), len(test_ary))
                ref_ary = ref_ary[:common_len]
                test_ary = test_ary[:common_len]

                error_is_small, error = check_result(test_ary, ref_ary)
                if not error_is_small:
                    raise AutomaticTestFailure(error)

                need_check = False

    events = []
    queue.finish()

    logger.info("%s: warmup done" % (test_entrypoint))

    logger.info("%s: timing run" % (test_entrypoint))

    timing_rounds = max(warmup_rounds, 1)

    while True:
        from time import time
        start_time = time()

        evt_start = cl.enqueue_marker(queue)

        for _i in range(timing_rounds):
            if not AUTO_TEST_SKIP_RUN:
                evt, _ = test_prog(queue, **args)
                events.append(evt)
            else:
                events.append(cl.enqueue_marker(queue))

        evt_end = cl.enqueue_marker(queue)

        queue.finish()
        stop_time = time()

        for evt in events:
            evt.wait()
        evt_start.wait()
        evt_end.wait()

        elapsed_event = (1e-9*events[-1].profile.END
                - 1e-9*events[0].profile.START) \
                / timing_rounds
        try:
            elapsed_event_marker = ((1e-9*evt_end.profile.START
                        - 1e-9*evt_start.profile.START)
                    / timing_rounds)
        except cl.RuntimeError:
            elapsed_event_marker = None

        elapsed_wall = (stop_time-start_time)/timing_rounds

        if elapsed_wall * timing_rounds < 0.3:
            timing_rounds *= 4
        else:
            break

    logger.info("%s: timing run done" % (test_entrypoint))

    rates = ""
    for cnt, lbl in zip(op_count, op_label):
        rates += " {:g} {}/s".format(cnt/elapsed_wall, lbl)

    if not quiet:
        def format_float_or_none(v):
            if v is None:
                return "<unavailable>"
            else:
                return "%g" % v

        print("elapsed: %s s event, %s s marker-event %s s wall "
                "(%d rounds)%s" % (
                    format_float_or_none(elapsed_event),
                    format_float_or_none(elapsed_event_marker),
                    format_float_or_none(elapsed_wall), timing_rounds, rates))

    if do_check:
        ref_rates = ""
        for cnt, lbl in zip(op_count, op_label):
            rates += " {:g} {}/s".format(cnt/elapsed_wall, lbl)

        if not quiet:
            print("elapsed: %s s event, %s s marker-event %s s wall "
                    "(%d rounds)%s" % (
                        format_float_or_none(elapsed_event),
                        format_float_or_none(elapsed_event_marker),
                        format_float_or_none(elapsed_wall), timing_rounds, rates))

        if do_check:
            ref_rates = ""
            for cnt, lbl in zip(op_count, op_label):
                ref_rates += " {:g} {}/s".format(cnt/ref_elapsed_event, lbl)
            if not quiet:
                print("ref: elapsed: {:g} s event, {:g} s wall{}".format(
                        ref_elapsed_event, ref_elapsed_wall, ref_rates))

    # }}}

    result_dict = {}
    result_dict["elapsed_event"] = elapsed_event
    result_dict["elapsed_event_marker"] = elapsed_event_marker
    result_dict["elapsed_wall"] = elapsed_wall
    result_dict["timing_rounds"] = timing_rounds

    if do_check:
        result_dict["ref_elapsed_event"] = ref_elapsed_event
        result_dict["ref_elapsed_wall"] = ref_elapsed_wall

    return result_dict


# }}}

# vim: foldmethod=marker



"""CUDA target independent of PyCUDA."""
from __future__ import annotations


__copyright__ = "Copyright (C) 2015 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from typing import TYPE_CHECKING, Sequence

import numpy as np

from cgen import Const, Declarator, Generable, Pointer
from pymbolic import var
from pytools import memoize_method

from loopy.diagnostic import LoopyError, LoopyTypeError
from loopy.kernel.array import ArrayBase, FixedStrideArrayDimTag, VectorArrayDimTag
from loopy.kernel.data import (
    AddressSpace,
    ArrayArg,
    ConstantArg,
    ImageArg,
    TemporaryVariable,
)
from loopy.kernel.function_interface import ScalarCallable
from loopy.target.c import CFamilyASTBuilder, CFamilyTarget
from loopy.target.c.codegen.expression import ExpressionToCExpressionMapper
from loopy.types import NumpyType


if TYPE_CHECKING:
    from loopy.codegen import CodeGenerationState
    from loopy.codegen.result import CodeGenerationResult


# {{{ vector types

class vec:  # noqa
    pass


def _create_vector_types():
    field_names = ["x", "y", "z", "w"]

    import sys
    if sys.maxsize <= 2**33:
        long_dtype = np.int32
        ulong_dtype = np.uint32
    else:
        long_dtype = np.int64
        ulong_dtype = np.uint64

    vec.types = {}
    vec.names_and_dtypes = []
    vec.type_to_scalar_and_count = {}

    for base_name, base_type, counts in [
            ("char", np.int8, [1, 2, 3, 4]),
            ("uchar", np.uint8, [1, 2, 3, 4]),
            ("short", np.int16, [1, 2, 3, 4]),
            ("ushort", np.uint16, [1, 2, 3, 4]),
            ("int", np.int32, [1, 2, 3, 4]),
            ("uint", np.uint32, [1, 2, 3, 4]),
            ("long", long_dtype, [1, 2, 3, 4]),
            ("ulong", ulong_dtype, [1, 2, 3, 4]),
            ("longlong", np.int64, [1, 2]),
            ("ulonglong", np.uint64, [1, 2]),
            ("float", np.float32, [1, 2, 3, 4]),
            ("double", np.float64, [1, 2]),
            ]:
        for count in counts:
            name = "%s%d" % (base_name, count)

            titles = field_names[:count]

            names = ["s%d" % i for i in range(count)]
            if len(titles) < len(names):
                titles.extend((len(names)-len(titles))*[None])

            try:
                dtype = np.dtype({
                    "names": names,
                    "formats": [base_type]*count,
                    "titles": titles})
            except NotImplementedError:
                try:
                    dtype = np.dtype([((n, title), base_type)
                                      for (n, title) in zip(names, titles)])
                except TypeError:
                    dtype = np.dtype([(n, base_type) for (n, title)
                                      in zip(names, titles)])

            setattr(vec, name, dtype)

            vec.names_and_dtypes.append((name, dtype))

            vec.types[np.dtype(base_type), count] = dtype
            vec.type_to_scalar_and_count[dtype] = np.dtype(base_type), count


_create_vector_types()


def _register_vector_types(dtype_registry):
    for name, dtype in vec.names_and_dtypes:
        dtype_registry.get_or_register_dtype(name, dtype)

# }}}


# {{{ function scoper

_CUDA_SPECIFIC_FUNCTIONS = {
        "rsqrt": 1,
        "atan2": 2,
        }


class CudaCallable(ScalarCallable):

    def cuda_with_types(self, arg_id_to_dtype, callables_table):

        name = self.name

        if name in _CUDA_SPECIFIC_FUNCTIONS:
            num_args = _CUDA_SPECIFIC_FUNCTIONS[name]

            # {{{ sanity checks

            for id, dtype in arg_id_to_dtype.items():
                if not -1 <= id < num_args:
                    raise LoopyError("%s can take only %d arguments." % (name,
                            num_args))

                if dtype is not None and dtype.kind == "c":
                    raise LoopyTypeError(
                        f"'{name}' does not support complex arguments.")

            # }}}

            for i in range(num_args):
                if i not in arg_id_to_dtype or arg_id_to_dtype[i] is None:
                    # the types provided aren't mature enough to specialize the
                    # callable
                    return (
                            self.copy(arg_id_to_dtype=arg_id_to_dtype),
                            callables_table)

            dtype = np.result_type(*[
                    dtype.numpy_dtype for id, dtype in arg_id_to_dtype.items()
                    if id >= 0])

            updated_arg_id_to_dtype = {id: NumpyType(dtype)
                    for id in range(-1, num_args)}

            return (
                    self.copy(name_in_target=name,
                        arg_id_to_dtype=updated_arg_id_to_dtype),
                    callables_table)

        if name == "dot":
            # CUDA dot function:
            # Performs dot product. Input types: vector and return type: scalar.
            for i in range(2):
                if i not in arg_id_to_dtype or arg_id_to_dtype[i] is None:
                    # the types provided aren't mature enough to specialize the
                    # callable
                    return (
                            self.copy(arg_id_to_dtype=arg_id_to_dtype),
                            callables_table)

            input_dtype = arg_id_to_dtype[0]

            scalar_dtype, _offset, _field_name = input_dtype.fields["x"]
            return_dtype = scalar_dtype
            return self.copy(arg_id_to_dtype={0: input_dtype, 1: input_dtype,
                                              -1: return_dtype})

        return (
                self.copy(arg_id_to_dtype=arg_id_to_dtype),
                callables_table)


def get_cuda_callables():
    cuda_func_ids = {"dot"} | set(_CUDA_SPECIFIC_FUNCTIONS)
    return {id_: CudaCallable(name=id_) for id_ in cuda_func_ids}

# }}}


# {{{ expression mapper

class ExpressionToCudaCExpressionMapper(ExpressionToCExpressionMapper):
    _GRID_AXES = "xyz"

    def map_group_hw_index(self, expr, type_context):
        return var(f"bIdx({self._GRID_AXES[expr.axis]})")

    def map_local_hw_index(self, expr, type_context):
        return var(f"tIdx({self._GRID_AXES[expr.axis]})")

# }}}


# {{{ target

class CudaTarget(CFamilyTarget):
    """A target for Nvidia's CUDA GPU programming language."""

    def __init__(self, extern_c=True):
        """
        :arg extern_c: If *True*, declare kernels using "extern C" to
            avoid name mangling.
        """
        self.extern_c = extern_c

        super().__init__()

    def split_kernel_at_global_barriers(self):
        return True

    def get_device_ast_builder(self):
        return CUDACASTBuilder(self)

    # {{{ types

    @memoize_method
    def get_dtype_registry(self):
        from loopy.target.c.compyte.dtypes import (
            DTypeRegistry,
            fill_registry_with_c_types,
        )

        result = DTypeRegistry()
        fill_registry_with_c_types(result, respect_windows=True)

        # no complex number support--needs PyOpenCLTarget

        _register_vector_types(result)

        return result

    def is_vector_dtype(self, dtype):
        return (isinstance(dtype, NumpyType)
                and dtype.numpy_dtype in list(vec.types.values()))

    def vector_dtype(self, base, count):
        return NumpyType(vec.types[base.numpy_dtype, count])

    # }}}

# }}}


# {{{ preamable generator

def cuda_preamble_generator(preamble_info):
    from loopy.types import AtomicNumpyType
    seen_64_bit_atomics = any(
            isinstance(dtype, AtomicNumpyType) and dtype.numpy_dtype.itemsize == 8
            for dtype in preamble_info.seen_atomic_dtypes)

    if seen_64_bit_atomics:
        # Source:
        # docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#atomic-functions
        yield ("00_enable_64bit_atomics", """
            #if __CUDA_ARCH__ < 600
            __device__ double atomicAdd(double* address, double val)
            {
                unsigned long long int* address_as_ull =
                                          (unsigned long long int*)address;
                unsigned long long int old = *address_as_ull, assumed;

                do {
                    assumed = old;
                    old = atomicCAS(address_as_ull, assumed,
                                    __double_as_longlong(val +
                                           __longlong_as_double(assumed)));

                } while (assumed != old);

                return __longlong_as_double(old);
            }
            #endif
            """)

    from loopy.tools import remove_common_indentation
    kernel = preamble_info.kernel
    idx_ctype = kernel.target.dtype_to_typename(kernel.index_dtype)
    yield ("00_declare_gid_lid",
           remove_common_indentation(f"""
                #define bIdx(N) (({idx_ctype}) blockIdx.N)
                #define tIdx(N) (({idx_ctype}) threadIdx.N)
           """))

# }}}


# {{{ ast builder

class CUDACASTBuilder(CFamilyASTBuilder):

    preamble_function_qualifier = "inline __device__"

    # {{{ library

    @property
    def known_callables(self):
        callables = super().known_callables
        callables.update(get_cuda_callables())
        return callables

    # }}}

    # {{{ top-level codegen

    def get_function_declaration(
            self, codegen_state: CodeGenerationState,
            codegen_result: CodeGenerationResult, schedule_index: int
            ) -> tuple[Sequence[tuple[str, str]], Generable]:
        preambles, fdecl = super().get_function_declaration(
                codegen_state, codegen_result, schedule_index)

        from loopy.target.c import FunctionDeclarationWrapper
        assert isinstance(fdecl, FunctionDeclarationWrapper)
        fdecl = fdecl.subdecl

        from cgen.cuda import CudaGlobal, CudaLaunchBounds
        fdecl = CudaGlobal(fdecl)

        if self.target.extern_c:
            from cgen import Extern
            fdecl = Extern("C", fdecl)

        from loopy.schedule import get_insn_ids_for_block_at
        assert codegen_state.kernel.linearization is not None
        _, local_grid_size = \
                codegen_state.kernel.get_grid_sizes_for_insn_ids_as_exprs(
                        get_insn_ids_for_block_at(
                            codegen_state.kernel.linearization, schedule_index),
                        codegen_state.callables_table)

        from loopy.symbolic import get_dependencies
        if not get_dependencies(local_grid_size):
            # Sizes can't have parameter dependencies if they are
            # to be used in static thread block size.
            from pytools import product
            nthreads = product(local_grid_size)

            fdecl = CudaLaunchBounds(nthreads, fdecl)

        return preambles, FunctionDeclarationWrapper(fdecl)

    def preamble_generators(self):

        return (
                [*super().preamble_generators(), cuda_preamble_generator])

    # }}}

    # {{{ code generation guts

    def get_expression_to_c_expression_mapper(self, codegen_state):
        return ExpressionToCudaCExpressionMapper(codegen_state)

    _VEC_AXES = "xyzw"

    def add_vector_access(self, access_expr, index):
        return access_expr.attr(self._VEC_AXES[index])

    def emit_barrier(self, synchronization_kind, mem_kind, comment):
        """
        :arg kind: ``"local"`` or ``"global"``
        :arg memkind: unused
        :return: a :class:`loopy.codegen.GeneratedInstruction`.
        """
        if synchronization_kind == "local":
            if comment:
                comment = " /* %s */" % comment

            from cgen import Statement
            return Statement("__syncthreads()%s" % comment)
        elif synchronization_kind == "global":
            raise LoopyError("CUDA does not have global barriers")
        else:
            raise LoopyError("unknown barrier kind")

    # }}}

    # {{{ declarators

    def wrap_decl_for_address_space(
            self, decl: Declarator, address_space: AddressSpace) -> Declarator:
        from cgen.cuda import CudaGlobal, CudaShared
        if address_space == AddressSpace.GLOBAL:
            return CudaGlobal(decl)
        if address_space == AddressSpace.LOCAL:
            return CudaShared(decl)
        elif address_space == AddressSpace.PRIVATE:
            return decl
        else:
            raise ValueError("unexpected address_space: %s"
                    % address_space)

    def wrap_global_constant(self, decl: Declarator) -> Declarator:
        from cgen.cuda import CudaConstant, CudaGlobal
        assert isinstance(decl, CudaGlobal)
        decl = decl.subdecl
        return CudaConstant(decl)

    # duplicated in OpenCL, update there if updating here
    def get_array_base_declarator(self, ary: ArrayBase) -> Declarator:
        dtype = ary.dtype

        vec_size = ary.vector_length()
        if vec_size > 1:
            dtype = self.target.vector_dtype(dtype, vec_size)

        if ary.dim_tags:
            for dim_tag in ary.dim_tags:
                if isinstance(dim_tag, (FixedStrideArrayDimTag, VectorArrayDimTag)):
                    # we're OK with those
                    pass

                else:
                    raise NotImplementedError(
                        f"{type(self).__name__} does not understand axis tag "
                        f"'{type(dim_tag)}.")

        from loopy.target.c import POD
        return POD(self, dtype, ary.name)

    def get_array_arg_declarator(
            self, arg: ArrayArg, is_written: bool) -> Declarator:
        from cgen.cuda import CudaRestrictPointer
        arg_decl: Declarator = CudaRestrictPointer(
                    self.get_array_base_declarator(arg))

        if not is_written:
            arg_decl = Const(arg_decl)

        return arg_decl

    def get_constant_arg_declarator(self, arg: ConstantArg) -> Declarator:
        from cgen import RestrictPointer
        from cgen.cuda import CudaConstant

        # constant *is* an address space as far as CUDA is concerned, do not re-wrap
        return CudaConstant(RestrictPointer(self.get_array_base_declarator(arg)))

    def get_image_arg_declarator(
            self, arg: ImageArg, is_written: bool) -> Declarator:
        raise NotImplementedError("not yet: texture arguments in CUDA")

    def emit_temp_var_decl_for_tv_with_base_storage(self,
                                                    codegen_state: CodeGenerationState,
                                                    tv: TemporaryVariable) -> Generable:
        from cgen import Initializer

        from loopy.target.c import POD, _ConstPointer, _ConstRestrictPointer

        assert tv.base_storage is not None
        ecm = codegen_state.expression_to_code_mapper

        cast_decl: Declarator = POD(self, tv.dtype, "")
        temp_var_decl: Declarator = POD(self, tv.dtype, tv.name)

        if tv._base_storage_access_may_be_aliasing:
            ptrtype: type[Pointer] = _ConstPointer
        else:
            # The 'restrict' part of this is a complete lie--of course
            # all these temporaries are aliased. But we're promising to
            # not use them to shovel data from one representation to the
            # other. That counts, right?
            ptrtype = _ConstRestrictPointer

        cast_decl = ptrtype(cast_decl)
        temp_var_decl = ptrtype(temp_var_decl)

        cast_tp, cast_d = cast_decl.get_decl_pair()
        return Initializer(
            temp_var_decl,
            "({} {}) ({} + {})".format(
                " ".join(cast_tp), cast_d, tv.base_storage, ecm(tv.offset)
            ),
        )

    # }}}

    # {{{ atomics

    def emit_atomic_update(self, codegen_state, lhs_atomicity, lhs_var,
            lhs_expr, rhs_expr, lhs_dtype, rhs_type_context):

        from cgen import Statement
        from pymbolic.mapper.stringifier import PREC_NONE
        from pymbolic.primitives import Sum

        if isinstance(lhs_dtype, NumpyType) and lhs_dtype.numpy_dtype in [
                np.int32, np.int64, np.float32, np.float64]:
            # atomicAdd
            if isinstance(rhs_expr, Sum):
                ecm = self.get_expression_to_code_mapper(codegen_state)

                new_rhs_expr = Sum(tuple(c for c in rhs_expr.children
                                         if c != lhs_expr))
                lhs_expr_code = ecm(lhs_expr)
                rhs_expr_code = ecm(new_rhs_expr)

                return Statement("atomicAdd(&{}, {})".format(
                    lhs_expr_code, rhs_expr_code))
            else:
                from cgen import Assign, Block, DoWhile

                from loopy.target.c import POD
                old_val_var = codegen_state.var_name_generator("loopy_old_val")
                new_val_var = codegen_state.var_name_generator("loopy_new_val")

                from loopy.kernel.data import TemporaryVariable
                ecm = codegen_state.expression_to_code_mapper.with_assignments(
                        {
                            old_val_var: TemporaryVariable(old_val_var, lhs_dtype),
                            new_val_var: TemporaryVariable(new_val_var, lhs_dtype),
                            })

                lhs_expr_code = ecm(lhs_expr, prec=PREC_NONE, type_context=None)

                from pymbolic import var
                from pymbolic.mapper.substitutor import make_subst_func

                from loopy.symbolic import SubstitutionMapper

                subst = SubstitutionMapper(
                        make_subst_func({lhs_expr: var(old_val_var)}))
                rhs_expr_code = ecm(subst(rhs_expr), prec=PREC_NONE,
                        type_context=rhs_type_context,
                        needed_dtype=lhs_dtype)

                cast_str = ""
                old_val = old_val_var
                new_val = new_val_var

                if lhs_dtype.numpy_dtype.kind == "f":
                    if lhs_dtype.numpy_dtype == np.float32:
                        ctype = "int"
                    elif lhs_dtype.numpy_dtype == np.float64:
                        ctype = "long"
                    else:
                        raise AssertionError()

                    old_val = "*(%s *) &" % ctype + old_val
                    new_val = "*(%s *) &" % ctype + new_val
                    cast_str = "(%s *) " % (ctype)

                return Block([
                    POD(self, NumpyType(lhs_dtype.dtype),
                        old_val_var),
                    POD(self, NumpyType(lhs_dtype.dtype),
                        new_val_var),
                    DoWhile(
                        "atomicCAS("
                        "%(cast_str)s&(%(lhs_expr)s), "
                        "%(old_val)s, "
                        "%(new_val)s"
                        ") != %(old_val)s"
                        % {
                            "cast_str": cast_str,
                            "lhs_expr": lhs_expr_code,
                            "old_val": old_val,
                            "new_val": new_val,
                            },
                        Block([
                            Assign(old_val_var, lhs_expr_code),
                            Assign(new_val_var, rhs_expr_code),
                            ])
                        )
                    ])
        else:
            raise NotImplementedError("atomic update for '%s'" % lhs_dtype)

    # }}}

# }}}

# vim: foldmethod=marker
